# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}
from __future__ import annotations

import base64
import logging
import mimetypes
import os
from pathlib import Path
import re
from pprint import pformat
from urllib.parse import urlparse, parse_qs
import zlib
from collections import defaultdict

import gevent
import gevent.pywsgi
import jwt
from cryptography.hazmat.primitives import serialization
from gevent import Greenlet
from jinja2 import Environment, FileSystemLoader, select_autoescape
from volttron.types import ServiceInterface
from volttron.types.server_config import ServerConfig
from werkzeug import Response

from ws4py.server.geventserver import WSGIServer

from .admin_endpoints import AdminEndpoints
from .vui_endpoints import VUIEndpoints
from .authenticate_endpoint import AuthenticateEndpoints
from .csr_endpoints import CSREndpoints
from .webapp import WebApplicationWrapper
from volttron.client.known_identities import CONTROL, VOLTTRON_CENTRAL, AUTH
from volttron.utils.context import ClientContext

from volttron.services.auth import AuthEntry, AuthFile, AuthFileEntryAlreadyExists
from volttron.utils.certs import Certs, CertWrapper
from volttron.utils.jsonrpc import (json_result, json_validate_request, INVALID_REQUEST, METHOD_NOT_FOUND,
                                    UNHANDLED_EXCEPTION, UNAUTHORIZED, UNAVAILABLE_PLATFORM, INVALID_PARAMS,
                                    UNAVAILABLE_AGENT, INTERNAL_ERROR, RemoteError)

from volttron.client.vip.agent import Agent, Core, RPC, Unreachable
from volttron.client.vip.agent.subsystems import query
from volttron.utils.keystore import encode_key
from volttron.utils import jsonapi, jsonrpc
from volttron.utils.network import is_ip_private
#from volttron.utils.rmq_config_params import RMQConfig

# must be after importing of utils which imports grequest.
import requests

_log = logging.getLogger(__name__)


class CouldNotRegister(Exception):
    pass


class DuplicateEndpointError(Exception):
    pass


__PACKAGE_DIR__ = os.path.dirname(os.path.abspath(__file__))
__TEMPLATE_DIR__ = os.path.join(__PACKAGE_DIR__, "templates")
__STATIC_DIR__ = os.path.join(__PACKAGE_DIR__, "static")


# Our admin interface will use Jinja2 templates based upon the above paths
# reference api for using Jinja2 http://jinja.pocoo.org/docs/2.10/api/
# Using the FileSystemLoader instead of the package loader in this case however.
tplenv = Environment(
    loader=FileSystemLoader(__TEMPLATE_DIR__),
    autoescape=select_autoescape(['html', 'xml'])
)


class PlatformWebService(ServiceInterface, Agent):
    """The service that is responsible for managing and serving registered pages

    Agents can register either a directory of files to serve or an rpc method
    that will be called during the request process.
    """

    def __init__(self, server_config: ServerConfig, bind_web_address,
                 web_ssl_key=None, web_ssl_cert=None, web_secret_key=None, **kwargs):
        """
        Initialize the configuration of the base web service integration within the platform.

        """
        self._web_secret_key = web_secret_key
        self._ssl_key = web_ssl_key
        self._ssl_cert = web_ssl_cert
        super(PlatformWebService, self).__init__(**kwargs)

        # no matter what we need to have a bind-web-address passed to us.
        if not bind_web_address:
            raise ValueError("bind-web-address not set in server_config.yml")

        _log.debug(pformat(server_config.opts.__dict__))
        self.bind_web_address = bind_web_address
        self.serverkey = server_config.opts.volttron_publickey
        self.instance_name = None
        self.registeredroutes = []
        self.peerroutes = defaultdict(list)
        self.pathroutes = defaultdict(list)
        # These will be used if set rather than the
        # any of the internal agent's certificates
        # self.web_ssl_key = web_ssl_key
        # self.web_ssl_cert = web_ssl_cert
        # self._web_secret_key = web_secret_key

        # Maps from endpoint to peer.
        self.endpoints = {}

        # self.volttron_central_address = volttron_central_address
        # self.volttron_central_rmq_address = volttron_central_rmq_address

        # If vc is this instance then make the vc address the same as
        # the web address.
        # if not self.volttron_central_address:
        #     self.volttron_central_address = bind_web_address

        # Initialize the mimetypes so that we can guess at the passed mimetype
        if not mimetypes.inited:
            mimetypes.init()

        self._certs = Certs()
        # noinspection PyTypeChecker
        self._csr_endpoints: CSREndpoints = None
        # noinspection PyTypeChecker
        self.appContainer: WebApplicationWrapper = None
        # noinspection PyTypeChecker
        self._server_greenlet: Greenlet = None
        # noinspection PyTypeChecker
        self._admin_endpoints: AdminEndpoints = None
        # noinspection PyTypeChecker
        self._vui_endpoints: VUIEndpoints = None

    @property
    def ssl_cert(self) -> str | None:
        """
        Web server ssl certificate path
        """
        return self._ssl_cert

    @property
    def ssl_key(self) -> str | None:
        """
        Web server ssl key path
        """
        return self._ssl_key

    # pylint: disable=unused-argument
    @Core.receiver('onsetup')
    def onsetup(self, sender, **kwargs):
        self.vip.rpc.export(self._auto_allow_csr, 'auto_allow_csr')
        self.vip.rpc.export(self._is_auto_allow_csr, 'is_auto_allow_csr')

    def _is_auto_allow_csr(self):
        return self._csr_endpoints.auto_allow_csr

    def _auto_allow_csr(self, auto_allow_csr):
        self._csr_endpoints.auto_allow_csr = auto_allow_csr

    def remove_unconnnected_routes(self):
        peers = self.vip.peerlist().get()

        for p in self.peerroutes:
            if p not in peers:
                del self.peerroutes[p]

    @RPC.export
    def get_user_claims(self, bearer):
        from ..web import get_user_claim_from_bearer
        if self.core.messagebus == 'rmq':
            claims = get_user_claim_from_bearer(bearer,
                                                tls_public_key=self._certs.get_cert_public_key(
                                                    ClientContext.get_fq_identity(self.core.identity)))
        elif self.ssl_cert is not None:
            claims = get_user_claim_from_bearer(bearer,
                                                tls_public_key=CertWrapper.get_cert_public_key(self.ssl_cert))
        elif self._web_secret_key is not None:
            claims = get_user_claim_from_bearer(bearer, web_secret_key=self._web_secret_key)

        else:
            raise ValueError("Configuration error secret key or web ssl cert must be not None.")

        return claims if claims.get('grant_type') == 'access_token' else {}

    @RPC.export
    def websocket_send(self, endpoint, message):
        _log.debug("Sending data to {} with message {}".format(endpoint,
                                                               message))
        self.appContainer.websocket_send(endpoint, message)

    @RPC.export
    def print_websocket_clients(self):
        _log.debug(self.appContainer.endpoint_clients)

    @RPC.export
    def get_bind_web_address(self):
        return self.bind_web_address

    @RPC.export
    def get_serverkey(self):
        return self.serverkey

    @RPC.export
    def get_volttron_central_address(self):
        """Return address of external Volttron Central

        Note: this only applies to Volttron Central agents that are
        running on a different platform.
        """
        return self.volttron_central_address

    @RPC.export
    def register_endpoint(self, endpoint, res_type):
        """
        RPC method to register a dynamic route.

        :param endpoint:
        :return:
        """
        # Get calling identity from whom the request came from
        identity = self.vip.rpc.context.vip_message.peer
        _log.debug('Registering route with endpoint: {}'.format(endpoint))
        _log.debug('Route is associated with peer: {}'.format(identity))

        if endpoint in self.endpoints:
            _log.error("Attempting to register an already existing endpoint.")
            _log.error("Ignoring registration.")
            raise DuplicateEndpointError(
                "Endpoint {} is already an endpoint".format(endpoint))

        self.endpoints[endpoint] = (identity, res_type)

    @RPC.export
    def register_agent_route(self, regex, fn):
        """ Register an agent route to an exported function.

        When a http request is executed and matches the passed regular
        expression then the function on peer is executed.
        """
        # Get calling identity from whom the request came from
        identity = self.vip.rpc.context.vip_message.peer

        _log.info(
            'Registering agent route expression: {} peer: {} function: {}'
                .format(regex, identity, fn))

        # TODO: inspect peer for function

        compiled = re.compile(regex)
        self.peerroutes[identity].append(compiled)
        self.registeredroutes.insert(0, (compiled, 'peer_route', (identity, fn)))

    @RPC.export
    def unregister_all_agent_routes(self):
        # Get calling identity from whom the request came from
        identity = self.vip.rpc.context.vip_message.peer

        _log.info('Unregistering agent routes for: {}'.format(identity))
        for regex in self.peerroutes[identity]:
            out = [cp for cp in self.registeredroutes if cp[0] != regex]
            self.registeredroutes = out
        del self.peerroutes[identity]
        for regex in self.pathroutes[identity]:
            out = [cp for cp in self.registeredroutes if cp[0] != regex]
            self.registeredroutes = out
        del self.pathroutes[identity]

        _log.debug(self.endpoints)
        endpoints = self.endpoints.copy()
        endpoints = {i:endpoints[i] for i in endpoints if endpoints[i][0] != identity}
        _log.debug(endpoints)
        self.endpoints = endpoints

    @RPC.export
    def register_path_route(self, regex, root_dir):
        # Get calling identity from whom the request came from
        identity = self.vip.rpc.context.vip_message.peer

        _log.info(f'Registering web path route from {identity} regex: {regex} dir: {root_dir}')

        compiled = re.compile(regex)
        self.pathroutes[identity].append(compiled)
        assert Path(root_dir).exists()
        # Make sure we resolve the root directory so its easier to check
        # later on.
        root_dir = str(Path(root_dir).resolve(root_dir))
        # in order for this agent to pass against the default route we want this
        # to be before the last route which will resolve to .*
        self.registeredroutes.insert(len(self.registeredroutes) - 1, (compiled, 'path', root_dir))

    @RPC.export
    def register_websocket(self, endpoint):
        # Get calling identity from whom the request came from
        identity = self.vip.rpc.context.vip_message.peer

        _log.debug('Caller identity: {}'.format(identity))
        _log.debug('REGISTERING ENDPOINT: {}'.format(endpoint))
        if self.appContainer:
            self.appContainer.create_ws_endpoint(endpoint, identity)
        else:
            _log.error('Attempting to register endpoint without web'
                       'subsystem initialized')
            raise AttributeError("self does not contain"
                                 " attribute appContainer")

    @RPC.export
    def unregister_websocket(self, endpoint):
        # Get calling identity from whom the request came from
        identity = self.vip.rpc.context.vip_message.peer

        _log.debug('Caller identity: {}'.format(identity))
        self.appContainer.destroy_ws_endpoint(endpoint)

    def _redirect_index(self, env, start_response, data=None):
        """ Redirect to the index page.
        @param env:
        @param start_response:
        @param data:
        @return:
        """
        start_response('302 Found', [('Location', '/index.html')])
        return [b'1']

    def _allow(self, environ, start_response, data=None):
        _log.info('Allowing new vc instance to connect to server.')
        jsondata = jsonapi.loads(data)
        json_validate_request(jsondata)

        assert jsondata.get('method') == 'allowvc'
        assert jsondata.get('params')

        params = jsondata.get('params')
        if isinstance(params, list):
            vcpublickey = params[0]
        else:
            vcpublickey = params.get('vcpublickey')

        assert vcpublickey
        assert len(vcpublickey) == 43

        authentry = {"credentials": vcpublickey, "identity": VOLTTRON_CENTRAL}
        try:
            self.vip.rpc.call(AUTH, "auth_file.add", authentry).get()
        except AuthFileEntryAlreadyExists:
            pass

        start_response('200 OK',
                       [('Content-Type', 'application/json')])
        return [jsonapi.dumpb(
            json_result(jsondata['id'], "Added")
        )]

    def _get_discovery(self, environ, start_response, data=None):
        q = query.Query(self.core)

        self.instance_name = q.query('instance-name').get(timeout=60)
        addreses = q.query('addresses').get(timeout=60)
        external_vip = None
        for x in addreses:
            try:
                if not is_ip_private(x):
                    external_vip = x
                    break
            except IndexError:
                pass

        return_dict = {}

        # Only send vip and serverkey if the platform has specified
        # a tcp address in the <VOLTTRON_HOME>/config or --vip-address command line argument.
        if external_vip and self.serverkey:
            return_dict['serverkey'] = encode_key(self.serverkey)
            return_dict['vip-address'] = external_vip
        elif not external_vip:
            _log.warning("There was no external vip-address specified in config file or command line.")

        if self.instance_name:
            return_dict['instance-name'] = self.instance_name

        # if self.core.messagebus == 'rmq':
        #     config = RMQConfig()
        #     rmq_address = None
        #     if config.is_ssl:
        #         rmq_address = "amqps://{host}:{port}/{vhost}".format(host=config.hostname, port=config.amqp_port_ssl,
        #                                                              vhost=config.virtual_host)
        #     else:
        #         rmq_address = "amqp://{host}:{port}/{vhost}".format(host=config.hostname, port=config.amqp_port,
        #                                                             vhost=config.virtual_host)
        #     return_dict['rmq-address'] = rmq_address
        #     return_dict['rmq-ca-cert'] = self._certs.cert(self._certs.root_ca_name).public_bytes(
        #         serialization.Encoding.PEM).decode("utf-8")
        return Response(jsonapi.dumps(return_dict), content_type="application/json")

    def app_routing(self, env, start_response):
        """
        The main routing function that maps the incoming request to a response.

        Depending on the registered routes map the request data onto an rpc
        function or a specific named file.
        """
        path_info = env['PATH_INFO']

        if path_info.startswith('/http://'):
            path_info = path_info[path_info.index('/', len('/http://')):]

        # only expose a partial list of the env variables to the registered
        # agents.
        envlist = ['HTTP_USER_AGENT', 'PATH_INFO', 'QUERY_STRING',
                   'REQUEST_METHOD', 'SERVER_PROTOCOL', 'REMOTE_ADDR',
                   'HTTP_ACCEPT_ENCODING', 'HTTP_COOKIE', 'CONTENT_TYPE',
                   'HTTP_AUTHORIZATION', 'SERVER_NAME', 'wsgi.url_scheme',
                   'HTTP_HOST']
        data = env['wsgi.input'].read().decode('utf-8')
        passenv = dict(
            (envlist[i], env[envlist[i]]) for i in range(0, len(envlist)) if envlist[i] in env.keys())

        _log.debug('path_info is: {}'.format(path_info))
        # Get the peer responsible for dealing with the endpoint.  If there
        # isn't a peer then fall back on the other methods of routing.
        (peer, res_type) = self.endpoints.get(path_info, (None, None))
        _log.debug('Peer path_info is associated with: {}'.format(peer))

        if self.is_json_content(env):
            data = jsonapi.loads(data)

        # Only if https available and rmq for the admin area.
        if env['wsgi.url_scheme'] == 'https' and self.core.messagebus == 'rmq':
            # Load the publickey that was used to sign the login message through the env
            # parameter so agents can use it to verify the Bearer has specific
            # jwt claims
            passenv['WEB_PUBLIC_KEY'] = env['WEB_PUBLIC_KEY'] = self._certs.get_cert_public_key(
                ClientContext.get_fq_identity(self.core.identity)).decode('utf-8')

        # if we have a peer then we expect to call that peer's web subsystem
        # callback to perform whatever is required of the method.
        if peer:
            _log.debug('Calling peer {} back with env={} data={}'.format(
                peer, passenv, data
            ))
            res = self.vip.rpc.call(peer, 'route.callback',
                                    passenv, data).get(timeout=60)

            if res_type == "jsonrpc":
                return self.create_response(res, start_response)
            elif res_type == "raw":
                return self.create_raw_response(res, start_response)

        env['JINJA2_TEMPLATE_ENV'] = tplenv

        # if ws4pi.socket is set then this connection is a web socket
        # and so we return the websocket response.

        if 'ws4py.socket' in env:
            return env['ws4py.socket'](env, start_response)

        for k, t, v in self.registeredroutes:
            if k.match(path_info):
                _log.debug("MATCHED:\npattern: {}, path_info: {}\n v: {}"
                           .format(k.pattern, path_info, v))
                _log.debug('registered route t is: {}'.format(t))
                if t == 'callable':  # Generally for locally called items.
                    # Changing signature of the "locally" called points to return
                    # a Response object. Our response object then will in turn
                    # be processed and the response will be written back to the
                    # calling client.
                    try:
                        retvalue = v(env, start_response, data)
                    except TypeError:
                        response = v(env, data)
                        return response(env, start_response)
                        # retvalue = self.process_response(start_response, v(env, data))

                    if isinstance(retvalue, Response):
                        return retvalue(env, start_response)
                    else:
                        return retvalue[0]

                elif t == 'peer_route':  # RPC calls from agents on the platform
                    _log.debug('Matched peer_route with pattern {}'.format(
                        k.pattern))
                    peer, fn = (v[0], v[1])
                    res = self.vip.rpc.call(peer, fn, passenv, data).get(
                        timeout=120)
                    _log.debug(res)
                    return self.create_response(res, start_response)

                elif t == 'path':  # File service from agents on the platform.
                    if path_info == '/':
                        return self._redirect_index(env, start_response)
                    server_path = v + path_info  # os.path.join(v, path_info)
                    server_path = str(Path(server_path).resolve())
                    _log.debug('Serverpath: {}'.format(server_path))
                    # protects against relative server traversal.
                    if not server_path.startswith(v):
                        start_response('403 Forbidden', [('Content-Type', 'text/html')])
                        return [b'<h1>403 Forbidden</h1>']
                    return self._sendfile(env, start_response, server_path)

        start_response('404 Not Found', [('Content-Type', 'text/html')])
        return [b'<h1>Not Found</h1>']

    def is_json_content(self, env):
        ct = env.get('CONTENT_TYPE')
        if ct is not None and 'application/json' in ct:
            return True
        return False

    def process_response(self, start_response, response):
        # if we are using the original response, then morph it into a werkzueg response.
        # response = PlatformWebService.convert_response_to_werkzueg(response)
        # return response()
        # process the response
        start_response(response.status, response.headers)

        if isinstance(response.content, str):
            return [response.content.encode('utf-8')]
        return [response.content]

    def create_raw_response(self, res, start_response):
        # If this is a tuple then we know we are going to have a response
        # and a headers portion of the data.
        if isinstance(res, tuple) or isinstance(res, list):
            if len(res) == 1:
                status, = res
                headers = ()
            elif len(res) == 2:
                headers = ()
                status, response = res
            elif len(res) == 3:
                status, response, headers = res
            else:
                raise Exception("Couldn't process raw response {}".format(res))
            start_response(status.encode('utf-8'), headers)
            return [base64.b64decode(response)]
        else:
            start_response("500 Programming Error",
                           [('Content-Type', 'text/html')])
            _log.error("Invalid length of response tuple (must be 1-3)")
            return [b'Invalid response tuple (must contain 1-3 elements)']

    def create_response(self, res, start_response):

        # Dictionaries are going to be treated as if they are meant to be json
        # serialized with Content-Type of application/json
        if isinstance(res, dict):
            # Note this is specific to volttron central agent and should
            # probably not be at this level of abstraction.
            _log.debug('res is a dictionary.')
            if 'error' in res.keys():
                if res['error']['code'] == UNAUTHORIZED:
                    start_response('401 Unauthorized', [
                        ('Content-Type', 'text/html')])
                    message = res['error']['message']
                    code = res['error']['code']
                    return ['<h1>{}</h1>\n<h2>CODE:{}</h2>'.format(message, code).encode('utf-8')]

            start_response('200 OK',
                           [('Content-Type', 'application/json')])
            return [jsonapi.dumpb(res)]
        elif isinstance(res, list):
            _log.debug('list implies [content, headers] or [status, content, headers]')
            if len(res) == 2:
                start_response('200 OK',
                               res[1])
                return res[0]
            elif len(res) == 3:
                start_response(res[0], res[2])
                if isinstance(res[1], str):
                    return [res[1].encode('utf-8')]
                return [res[1]]

        # If this is a tuple then we know we are going to have a response
        # and a headers portion of the data.
        if isinstance(res, tuple) or isinstance(res, list):
            if len(res) != 2:
                start_response("500 Programming Error",
                               [('Content-Type', 'text/html')])
                _log.error("Invalid length of response tuple (must be 2)")
                return [b'Invalid response tuple (must contain 2 elements)']

            response, headers = res
            header_dict = dict(headers)
            if header_dict.get('Content-Encoding', None) == 'gzip':
                gzip_compress = zlib.compressobj(9, zlib.DEFLATED,
                                                 zlib.MAX_WBITS | 16)
                data = gzip_compress.compress(response) + gzip_compress.flush()
                start_response('200 OK', headers)
                return [data]
            else:
                return [response]
        else:
            start_response('200 OK',
                           [('Content-Type', 'application/json')])
            return [jsonapi.dumpb(res)]

    def _sendfile(self, env, start_response, filename):
        from wsgiref.util import FileWrapper
        status = '200 OK'
        _log.debug('SENDING FILE: {}'.format(filename))
        guess = mimetypes.guess_type(filename)[0]
        _log.debug('MIME GUESS: {}'.format(guess))

        basename = os.path.dirname(filename)

        if not os.path.exists(basename):
            start_response('404 Not Found', [('Content-Type', 'text/html')])
            return [b'<h1>Not Found</h1>']
        elif not os.path.isfile(filename):
            start_response('404 Not Found', [('Content-Type', 'text/html')])
            return [b'<h1>Not Found</h1>']

        if not guess:
            guess = 'text/plain'

        response_headers = [
            ('Content-type', guess),
        ]
        start_response(status, response_headers)

        return FileWrapper(open(filename, 'rb'))

    def _to_jsonrpc_obj(self, jsonrpcstr):
        """ Convert data string into a JsonRpcData named tuple.

        :param object data: Either a string or a dictionary representing a json document.
        """
        return jsonrpc.JsonRpcData.parse(jsonrpcstr)

    def jsonrpc(self, env, data):
        """ The main entry point for ^jsonrpc data

        This method will only accept rpcdata.  The first time this method
        is called, per session, it must be using get_authorization.  That
        will return a session token that must be included in every
        subsequent request.  The session is tied to the ip address
        of the caller.

        :param object env: Environment dictionary for the request.
        :param object data: The JSON-RPC 2.0 method to call.
        :return object: An JSON-RPC 2.0 response.
        """
        if env['REQUEST_METHOD'].upper() != 'POST':
            return Response(jsonapi.dumps(jsonrpc.json_error('NA', INVALID_REQUEST,
                                      'Invalid request method, only POST allowed')), content_type="application/json")

        try:
            rpcdata = self._to_jsonrpc_obj(data)
            _log.info('rpc method: {}'.format(rpcdata.method))

            # Authenticate rpc call
            if 'authentication' in rpcdata.params:
                if self.jsonrpc_verify_and_dispatch(rpcdata.params['authentication']):
                    del rpcdata.params['authentication']
                else:
                    return Response(jsonapi.dumps(jsonrpc.json_error(rpcdata.id, UNAUTHORIZED,
                                                           "Invalid username/password specified.")),
                                        content_type="application/json")
            else:
                return Response(jsonapi.dumps(jsonrpc.json_error(rpcdata.id, UNAUTHORIZED,
                                                                     "Authentication parameter missing.")),
                                    content_type="application/json")

            _log.debug('RPC METHOD IS: {}'.format(rpcdata.method))
            if not rpcdata.method:
                return Response(jsonapi.dumps(jsonrpc.json_error(
                    'NA', INVALID_REQUEST, 'Invalid rpc data {}'.format(data))), content_type="application/json")
            else:
                if rpcdata.params:
                    result_or_error = self.vip.rpc(rpcdata.id, rpcdata.method, **rpcdata.params).get()
                else:
                    result_or_error = self.vip.rpc(rpcdata.id, rpcdata.method).get()

        except AssertionError:
            return Response(jsonapi.dumps(jsonrpc.json_error(
                'NA', INVALID_REQUEST, 'Invalid rpc data {}'.format(data))), content_type="application/json")
        except Unreachable:
            return Response(
                jsonapi.dumps(jsonrpc.json_error(rpcdata.id, UNAVAILABLE_PLATFORM,
                                                 "Couldn't reach platform with method {} params: {}".format(
                                                     rpcdata.method, rpcdata.params))),
                content_type="application/json")
        except Exception as e:

            return Response(jsonapi.dumps(jsonrpc.json_error('NA', UNHANDLED_EXCEPTION, e)),
                                content_type="application/json")

        return Response(jsonapi.dumps(self._get_jsonrpc_response(rpcdata.id, result_or_error)),
                            content_type="application/json")

    def _get_jsonrpc_response(self, id, result_or_error):
        """ Wrap the response in either a json-rpc error or result.

        :param id:
        :param result_or_error:
        :return:
        """
        if isinstance(result_or_error, dict):
            if 'jsonrpc' in result_or_error:
                return result_or_error

        if result_or_error is not None and isinstance(result_or_error, dict):
            if 'error' in result_or_error:
                error = result_or_error['error']
                _log.debug("RPC RESPONSE ERROR: {}".format(error))
                return jsonrpc.json_error(id, error['code'], error['message'])
        return jsonrpc.json_result(id, result_or_error)

    def jsonrpc_verify_and_dispatch(self, authentication):
        """ Verify that the user is an admin

        :param authentication: authentication generated by successful authentication
        :return: Boolean
        """
        from ..web import NotAuthorized
        try:
            claims = self.get_user_claims(authentication)
        except NotAuthorized:
            _log.error("Unauthorized user attempted to connect to platform.")
            return False
        except jwt.ExpiredSignatureError:
            _log.error("User attempted to connect with an expired signature.")
            return False

        return True



    @Core.receiver('onstart')
    def startupagent(self, sender, **kwargs):

        from urllib.parse import urlparse
        parsed = urlparse(self.bind_web_address)

        ssl_key = self._ssl_key
        ssl_cert = self._ssl_cert
        rpc_caller = self.vip.rpc
        if parsed.scheme == 'https':
            # Admin interface is only availble to rmq at present.
            if self.core.messagebus == 'rmq':
                self._admin_endpoints = AdminEndpoints(rmq_mgmt=self.core.rmq_mgmt,
                                                       ssl_public_key=self._certs.get_cert_public_key(
                                                           ClientContext.get_fq_identity(self.core.identity)),
                                                       rpc_caller=rpc_caller)
            if ssl_key is None or ssl_cert is None:
                # Because the  platform.web service certificate is a client to rabbitmq we
                # can't use it directly therefore we use the -server on the file to specify
                # the server based file.
                base_filename = ClientContext.get_fq_identity(self.core.identity) + "-server"
                _log.debug(f'############### BASE FILENAME: {base_filename} #########################')
                ssl_cert = self._certs.cert_file(base_filename)
                _log.debug(f'SSL CERT: {ssl_cert}')
                ssl_key = self._certs.private_key_file(base_filename)
                _log.debug(f'SSL KEY: {ssl_key}')

                if not os.path.isfile(ssl_cert) or not os.path.isfile(ssl_key):
                    self._certs.create_signed_cert_files(base_filename, cert_type='server')

            if ssl_key is not None and ssl_cert is not None and self._admin_endpoints is None:
                self._admin_endpoints = AdminEndpoints(ssl_public_key=CertWrapper.get_cert_public_key(ssl_cert),
                                                       rpc_caller=rpc_caller)
        else:
            self._admin_endpoints = AdminEndpoints(rpc_caller=rpc_caller)

        hostname = parsed.hostname
        port = parsed.port

        _log.info('Starting web server binding to {}://{}:{}.'.format(parsed.scheme,
                                                                      hostname, port))
        # Handle the platform.web routes here.
        self.registeredroutes.append((re.compile('^/discovery/$'), 'callable', self._get_discovery))
        self.registeredroutes.append((re.compile('^/discovery/allow$'), 'callable', self._allow))
        self.registeredroutes.append((re.compile(r'/gs'), 'callable', self.jsonrpc))
        # these routes are only available for rmq based message bus
        # at present.
        if self.core.messagebus == 'rmq':
            # We need reference to the object so we can change the behavior of
            # whether or not to have auto certs be created or not.
            self._csr_endpoints = CSREndpoints(self.core)
            for rt in self._csr_endpoints.get_routes():
                self.registeredroutes.append(rt)

        # Register the admin endpoints regardless of whether there is an ssl context
        # or not.
        for rt in self._admin_endpoints.get_routes():
            self.registeredroutes.append(rt)

        # Register VUI endpoints:
        self._vui_endpoints = VUIEndpoints(self)
        self.registeredroutes.extend(self._vui_endpoints.get_routes())

        # Allow authentication endpoint from any https connection
        if parsed.scheme == 'https':
            if self.core.messagebus == 'rmq':
                ssl_private_key = self._certs.get_pk_bytes(ClientContext.get_fq_identity(self.core.identity))
                ssl_public_key = self._certs.get_cert_public_key(ClientContext.get_fq_identity(self.core.identity))
            else:
                ssl_private_key = CertWrapper.get_private_key(ssl_key)
                ssl_public_key = CertWrapper.get_cert_public_key(self.web_ssl_cert)
            for rt in AuthenticateEndpoints(tls_private_key=ssl_private_key, tls_public_key=ssl_public_key).get_routes():
                self.registeredroutes.append(rt)
        else:
            # We don't have a private ssl key if we aren't using ssl.
            for rt in AuthenticateEndpoints(web_secret_key=self._web_secret_key).get_routes():
                self.registeredroutes.append(rt)

        static_dir = os.path.join(os.path.dirname(__file__), "static")
        self.registeredroutes.append((re.compile('^/.*$'), 'path', static_dir))

        port = int(port)

        self.appContainer = WebApplicationWrapper(self, hostname, port)
        if ssl_key and ssl_cert:
            svr = WSGIServer((hostname, port), self.appContainer,
                             certfile=ssl_cert,
                             keyfile=ssl_key)
        else:
            svr = WSGIServer((hostname, port), self.appContainer)
        self._server_greenlet = gevent.spawn(svr.serve_forever)

    def _authenticate_route(self, env, start_response, data):
        scheme = env.get('wsgi.url_scheme')

        if scheme != 'https':
            _log.warning("Authentication should be through https")
            start_response("401 Unauthorized", [('Content-Type', 'text/html')])
            return "<html><body><h1>401 Unauthorized</h1></body></html>"

        from pprint import pprint
        pprint(env)

        import jwt

        jwt.encode()

    @Core.receiver('onstop')
    def onstop(self, sender, **kwargs):
        _log.debug("Stopping web agent.")
        if not self._server_greenlet.dead:
            self._server_greenlet.join(timeout=10)



import logging

from ws4py.server.wsgiutils import WebSocketWSGIApplication

from .websocket import VolttronWebSocket

_log = logging.getLogger(__name__)


class WebApplicationWrapper(object):
    """ A container class that will hold all of the applications registered
    with it.  The class provides a container for managing the routing of
    websocket, static content, and rpc function calls.
    """
    def __init__(self, platformweb, host, port):
        self.platformweb = platformweb
        self.port = port
        self.host = host
        self.ws = WebSocketWSGIApplication(handler_cls=VolttronWebSocket)
        self.clients = []
        self.endpoint_clients = {}
        self._wsregistry = {}

    def __call__(self, environ, start_response):
        """
        Good ol' WSGI application. This is a simple demo
        so I tried to stay away from dependencies.
        """
        if environ['PATH_INFO'] == '/favicon.ico':
            return self.favicon(environ, start_response)

        path = environ['PATH_INFO']
        if path in self._wsregistry:
            environ['ws4py.app'] = self
            environ['identity'] = self._wsregistry[environ['PATH_INFO']]
            return self.ws(environ, start_response)

        return self.platformweb.app_routing(environ, start_response)

    def favicon(self, environ, start_response):
        """
        Don't care about favicon, let's send nothing.
        """
        status = '200 OK'
        headers = [('Content-type', 'text/plain')]
        start_response(status, headers)
        return ""

    def client_opened(self, client, endpoint, identity):

        ip = client.environ['REMOTE_ADDR']
        should_open = self.platformweb.vip.rpc.call(identity, 'client.opened',
                                                  ip, endpoint)
        if not should_open:
            _log.error("Authentication failure, closing websocket.")
            client.close(reason='Authentication failure!')
            return

        # In order to get into endpoint_clients create_ws must be called.
        if endpoint not in self.endpoint_clients:
            _log.error('Unknown endpoint detected: {}'.format(endpoint))
            client.close(reason="Unknown endpoint! {}".format(endpoint))
            return

        if (identity, client) in  self.endpoint_clients[endpoint]:
            _log.debug("IDENTITY,CLIENT: {} already in endpoint set".format(identity))
        else:
            _log.debug("IDENTITY,CLIENT: {} added to endpoint set".format(identity))
            self.endpoint_clients[endpoint].add((identity, client))

    def client_received(self, endpoint, message):
        clients = self.endpoint_clients.get(endpoint, [])
        for identity, _ in clients:
            self.platformweb.vip.rpc.call(identity, 'client.message',
                                        str(endpoint), str(message))

    def client_closed(self, client, endpoint, identity,
                      reason="Client left without proper explanation"):

        client_set = self.endpoint_clients.get(endpoint, set())

        try:
            key = (identity, client)
            client_set.remove(key)
        except KeyError:
            pass
        else:
            self.platformweb.vip.rpc.call(identity, 'client.closed', endpoint)

    def create_ws_endpoint(self, endpoint, identity):
        if endpoint not in self.endpoint_clients:
            self.endpoint_clients[endpoint] = set()
        self._wsregistry[endpoint] = identity

    def destroy_ws_endpoint(self, endpoint):
        clients = self.endpoint_clients.get(endpoint, [])
        for identity, client in clients:
            client.close(reason="Endpoint closed.")
        try:
            del self.endpoint_clients[endpoint]
        except KeyError:
            pass

    def websocket_send(self, endpoint, message):
        _log.debug('Sending message to clients!')
        clients = self.endpoint_clients.get(endpoint, [])
        if not clients:
            _log.warning("There were no clients for endpoint {}".format(endpoint))
        for c in clients:
            identity, client = c
            _log.debug('Sending endpoint&&message {}&&{}'.format(
                endpoint, message))
            client.send(message)




import logging
import requests
from urllib.parse import urlparse, urljoin

from volttron.utils import jsonapi
from volttron.utils.certs import Certs

_log = logging.getLogger(__name__)


class DiscoveryError(Exception):
    """ Raised when a different volttron central tries to register.
    """
    pass


class DiscoveryInfo(object):
    """ A DiscoveryInfo class.

    The DiscoveryInfo class provides a wrapper around the return values from
    a call to the /discovery/ endpoint of the `volttron.services.web.
    """

    def __init__(self, **kwargs):

        self.discovery_address = kwargs.pop('discovery_address')
        self.vip_address = kwargs.pop('vip-address', None)
        self.serverkey = kwargs.pop('serverkey', None)
        self.instance_name = kwargs.pop('instance-name')
        try:
            self.rmq_address = kwargs.pop('rmq-address')
        except KeyError:
            self.messagebus_type = 'zmq'
        else:
            self.messagebus_type = 'rmq'
        try:
            self.rmq_ca_cert = kwargs.pop('rmq-ca-cert')
        except KeyError:
            self.messagebus_type = 'zmq'
        else:
            self.messagebus_type = 'rmq'
        self.certs = Certs()

        assert len(kwargs) == 0

    @staticmethod
    def request_discovery_info(web_address):
        """  Construct a `DiscoveryInfo` object.

        Requests a response from discovery_address and constructs a
        `DiscoveryInfo` object with the returned json.

        :param web_address: An http(s) address with volttron running.
        :return:
        """

        try:
            parsed = urlparse(web_address)

            assert parsed.scheme
            assert not parsed.path

            real_url = urljoin(web_address, "/discovery/")
            _log.info('Connecting to: {}'.format(real_url))
            response = requests.get(real_url, verify=False)

            if not response.ok:
                raise DiscoveryError(
                    "Invalid discovery response from {}".format(real_url)
                )
        except AttributeError as e:
            raise DiscoveryError(
                "Invalid web_address passed {}"
                .format(web_address)
            )
        except requests.exceptions.RequestException as e:
            raise DiscoveryError(
                "Connection to {} not available".format(real_url)
            )
        except Exception as e:
            raise DiscoveryError("Unhandled exception {}".format(e))

        return DiscoveryInfo(
            discovery_address=web_address, **(response.json()))

    def __str__(self):
        dk = {
            'discovery_address': self.discovery_address,
            'instance_name': self.instance_name,
            'messagebus_type': self.messagebus_type
        }
        if self.messagebus_type == 'rmq':
            dk['rmq_address'] = self.rmq_address
            dk['rmq_ca_cert'] = self.rmq_ca_cert

        if self.vip_address:
            dk['vip_address'] = self.vip_address
            dk['serverkey'] = self.serverkey

        return jsonapi.dumps(dk)



import json
from weakref import WeakValueDictionary
from collections import defaultdict

from .websocket import VolttronWebSocket
from ws4py.server.wsgiutils import WebSocketWSGIApplication
from ws4py.websocket import WebSocket, EchoWebSocket
import logging

_log = logging.getLogger()


class VUIPubsubManager:
    def __init__(self, agent):
        self._agent = agent
        self.subscription_websockets = WeakValueDictionary() # Websockets for all topics with current subscriptions.
        self.publication_websockets = {}  # Websockets for all topics with current publication queue.
        self.user_websockets = defaultdict(dict)  # References to all websockets for each user access_token.

    def get_socket_routes(self, access_token, topic=None):
        _log.debug('In get_socket_routes. User_websockets is: ')
        _log.debug(self.user_websockets)
        _log.debug('In get_socket_routes. subscription_websockets is: ')
        _log.debug(self.subscription_websockets)
        return {t: str(w) for t, w in self.user_websockets[access_token].items()}

    def open_subscription_socket(self, access_token, topic):
        _log.debug('In open_subscription_socket:')
        _log.debug((f'access_token: {access_token}'))
        _log.debug(f'topic: {topic}')
        ws = self.subscription_websockets.get(topic)
        _log.debug(f'subscription_websockets has: {ws}')
        if not ws:
            _log.debug("Didn't find ws, so creating.")
            ws = self._create_websocket(topic, access_token)
            _log.debug(f'Now we have ws: {ws}')
        # TODO: Create process for checking that tokens are still valid.
        else:
            _log.debug('We had ws, so now in else block.')
            self.user_websockets[access_token][topic] = ws
            _log.debug('reached end, returning ws.')
        return ws

    def close_socket(self, access_token, topic=None):
        # TODO: Cancel subscription to a topic (or all topics for ../pubsub).
        # TODO: Remove any topic expiry monitors.
        # TODO: Close websocket if there are no additional subscribers.
        pass

    def publish(self, topic, headers, message):
        subscriber_count = self._agent.vip.pubsub.publish('pubsub', topic, headers=headers, message=message).get(timeout=5)
        return {'number_of_subscribers': subscriber_count}

    def _create_websocket(self, topic, access_token):
        ws_app = WebSocketWSGIApplication(handler_cls=VUIWebSocket)
        self.subscription_websockets[topic] = ws_app
        self.user_websockets[access_token][topic] = self.subscription_websockets[topic]
        return ws_app

    def client_opened(self, ws, topic, access_token):
        _log.debug(f'VUIPubsubManager: Subscribing to {topic}')
        self._agent.vip.pubsub.subscribe('pubsub', topic, ws.on_topic)
        self.user_websockets[access_token][topic] = ws

        # if topic not in self.subscription_websockets:
        #     err = f'Websocket opened for unintialized topic: {topic}'
        #     _log.error(err)
        #     ws.close(reason=err)
        #     return

        # TODO: Use this if we need to look over websockets connected to one subscription.
        # if ws in  self.subscription_websockets[topic]:
        #     _log.debug("IDENTITY,CLIENT: {} already in endpoint set".format(identity))
        # else:
        #     _log.debug("IDENTITY,CLIENT: {} added to endpoint set".format(identity))
        #     self.subscription_websockets[endpoint].add((identity, ws))
#
# class SubscriptionGroup:
#     def __init__(self, pubsub_interface, topic: str):
#         self.topic = topic
#         self.websockets = []
#         self.pubsub = pubsub_interface
#         self.pubsub.subscribe('pubsub', topic, self.on_publish)
#
#
#     def add(self, ws: WebSocket):
#         self.websockets.append(ws)
#
#     def on_publish(self, peer, sender, bus, topic, headers, message):
#         for ws in self.websockets:
#             if not ws.terminated:
#                 try:
#                     ws.send(message)
#                 except Exception as e:
#                     _log.warning(f'Error sending subscription data: {e}')
#
#     def remove(self, ):
#         pass


class VUIWebSocket(WebSocket):
    def __init__(self, *args, **kwargs):
        super(VUIWebSocket, self).__init__(*args, **kwargs)
        _log = logging.getLogger(self.__class__.__name__)

    def _get_topic(self):
        from volttron.services.web import get_bearer
        path_info = self.environ['PATH_INFO']
        topic = path_info.split('/pubsub/')[1]
        access_token = get_bearer(self.environ)
        return topic, access_token

    def opened(self):
        _log.info('Socket opened')
        app = self.environ['ws4py.app']
        topic, access_token = self._get_topic()
        app.client_opened(self, topic, access_token)

    def received_message(self, m):
        # self.clients is set from within the server
        # and holds the list of all connected servers
        # we can dispatch to
        # _log.debug('Socket received message: {}'.format(m))
        # app = self.environ['ws4py.app']
        # identity, endpoint = self._get_identity_and_endpoint()
        # app.client_received(endpoint, m)
        pass

    def closed(self, code, reason="A client left the room without a proper explanation."):
        # _log.info('Socket closed!')
        # app = self.environ.pop('ws4py.app')
        # identity, endpoint = self._get_identity_and_endpoint()
        # app.client_closed(self, endpoint, identity, reason)
        pass

    def on_topic(self, peer, sender, bus, topic, headers, message):
        _log.debug('VUIWebSocket: in _on_topic')
        _log.debug(f'topic is: {topic}')
        _log.debug(f'message is: {message}')
        if not self.terminated:
            try:
                self.send(json.dumps(message))
            except Exception as e:
                _log.warning(f'Error sending subscription data: {e}')



import logging
import os
import re
from urllib.parse import parse_qs
from datetime import datetime, timedelta
import json

import jwt
from jinja2 import Environment, FileSystemLoader, select_autoescape
from passlib.hash import argon2
from watchdog_gevent import Observer
from werkzeug import Response

from volttron.utils.context import ClientContext
from volttron.utils.filewatch import VolttronHomeFileReloader
from volttron.utils.persistance import PersistentDict

_log = logging.getLogger(__name__)

__PACKAGE_DIR__ = os.path.dirname(os.path.abspath(__file__))
__TEMPLATE_DIR__ = os.path.join(__PACKAGE_DIR__, "templates")
__STATIC_DIR__ = os.path.join(__PACKAGE_DIR__, "static")


# Our admin interface will use Jinja2 templates based upon the above paths
# reference api for using Jinja2 http://jinja.pocoo.org/docs/2.10/api/
# Using the FileSystemLoader instead of the package loader in this case however.
tplenv = Environment(
    loader=FileSystemLoader(__TEMPLATE_DIR__),
    autoescape=select_autoescape(['html', 'xml'])
)


class AuthenticateEndpoints(object):

    def __init__(self, tls_private_key=None, tls_public_key=None, web_secret_key=None):

        self.refresh_token_timeout = 240  # minutes before token expires. TODO: Should this be a setting somewhere?
        self.access_token_timeout = 15  # minutes before token expires. TODO: Should this be a setting somewhere?
        self._tls_private_key = tls_private_key
        self._tls_public_key = tls_public_key
        self._web_secret_key = web_secret_key
        if self._tls_private_key is None and self._web_secret_key is None:
            raise ValueError("Must have either ssl_private_key or web_secret_key specified!")
        if self._tls_private_key is not None and self._web_secret_key is not None:
            raise ValueError("Must use either ssl_private_key or web_secret_key not both!")
        self._userdict = None
        self.reload_userdict()
        self._observer = Observer()
        self._observer.schedule(
            VolttronHomeFileReloader("web-users.json", self.reload_userdict),
            ClientContext.get_volttron_home()
        )
        self._observer.start()

    def reload_userdict(self):
        webuserpath = os.path.join(ClientContext.get_volttron_home(), 'web-users.json')
        self._userdict = PersistentDict(webuserpath)

    def get_routes(self):
        """
        Returns a list of tuples with the routes for authentication.

        Tuple should have the following:

            - regular expression for calling the endpoint
            - 'callable' keyword specifying that a method is being specified
            - the method that should be used to call when the regular expression matches

        code:

            return [
                (re.compile('^/csr/request_new$'), 'callable', self._csr_request_new)
            ]

        :return:
        """
        return [
            (re.compile('^/authenticate'), 'callable', self.handle_authenticate)
        ]

    def handle_authenticate(self, env, data):
        """
        Callback for /authenticate endpoint.

        Routes request based on HTTP method and returns an application/json encoded token or error.

        :param env:
        :param data:
        :return: Response
        """
        method = env.get('REQUEST_METHOD')
        if method == 'POST':
            response = self.get_auth_tokens(env, data)
        elif method == 'PUT':
            response = self.renew_auth_token(env, data)
        elif method == 'DELETE':
            response = self.revoke_auth_token(env, data)
        else:
            error = f"/authenticate endpoint accepts only POST, PUT, or DELETE methods. Received: {method}"
            _log.warning(error)
            return Response(json.dumps({'error': error}), status=405, content_type='application/json')
        return response

    def get_auth_tokens(self, env, data):
        """
        Creates an authentication refresh and acccss tokens to be returned to the caller.  The
        response will be application/json encoded refresh and access tokens.  Data should contain:
        {
            "username": "<username>",
            "password": "<password>"
        }
        :param env:
        :param data:
        :return:
        """

        assert len(self._userdict) > 0, "No users in user dictionary, set the administrator password first!"

        if not isinstance(data, dict):
            _log.debug("data is not a dict, decoding")
            decoded = dict((k, v if len(v) > 1 else v[0])
                           for k, v in parse_qs(data).items())

            username = decoded.get('username')
            password = decoded.get('password')

        else:
            username = data.get('username')
            password = data.get('password')

        _log.debug("Username is: {}".format(username))

        error = ""
        if username is None:
            error += "Invalid username passed"
        if not password:
            error += "Invalid password passed"

        if error:
            _log.error("Invalid parameters passed: {}".format(error))
            return Response(json.dumps({'error': error}), status=401, content_type='application/json')

        user = self.__get_user(username, password)
        if user is None:
            _log.error("No matching user for passed username: {}".format(username))
            return Response(json.dumps({'error': 'Not Authorized'}), status='401', content_type='application/json')
        access_token, refresh_token = self._get_tokens(user)
        response = Response(json.dumps({"refresh_token": refresh_token, "access_token": access_token}),
                            content_type="application/json")
        return response

    def _get_tokens(self, claims):
        now = datetime.utcnow()
        claims['iat'] = now
        claims['nbf'] = now
        claims['exp'] = now + timedelta(minutes=self.access_token_timeout)
        claims['grant_type'] = 'access_token'
        algorithm = 'RS256' if self._tls_private_key is not None else 'HS256'
        encode_key = self._tls_private_key if algorithm == 'RS256' else self._web_secret_key
        access_token = jwt.encode(claims, encode_key, algorithm=algorithm)
        claims['exp'] = now + timedelta(minutes=self.refresh_token_timeout)
        claims['grant_type'] = 'refresh_token'
        refresh_token = jwt.encode(claims, encode_key, algorithm=algorithm)
        return access_token.decode('utf-8'), refresh_token.decode('utf8')

    def renew_auth_token(self, env, data):
        """
        Creates a new authentication access token to be returned to the caller.  The
        response will be an application/json encoded access token.  Request should contain:
            • Content Type: application/json
            • Authorization: BEARER <jwt_refresh_token>
            • Body (optional):
                {
                "current_access_token": "<jwt_access_token>"
                }

        :param env:
        :param data:
        :return:
        """
        current_access_token = data.get('current_access_token')
        from ..web import get_bearer, get_user_claim_from_bearer, NotAuthorized
        try:
            current_refresh_token = get_bearer(env)
            claims = get_user_claim_from_bearer(current_refresh_token, web_secret_key=self._web_secret_key,
                                                tls_public_key=self._tls_public_key)
        except NotAuthorized:
            _log.error("Unauthorized user attempted to connect to {}".format(env.get('PATH_INFO')))
            return Response(json.dumps({'error': 'Not Authorized'}), status=401, content_type='application/json')

        except jwt.ExpiredSignatureError:
            _log.error("User attempted to connect to {} with an expired signature".format(env.get('PATH_INFO')))
            return Response(json.dumps({'error': 'Not Authorized'}), status=401, content_type='application/json')

        if claims.get('grant_type') != 'refresh_token' or not claims.get('groups'):
            return Response(json.dumps({'error': 'Not Authorized'}), status=401, content_type='application/json')
        else:
            # TODO: Consider blacklisting and reissuing refresh tokens also when used.
            new_access_token, _ = self._get_tokens(claims)
            if current_access_token:
                pass  # TODO: keep current subscriptions? blacklist old token?
            return Response(json.dumps({"access_token": new_access_token}), content_type="application/json")

    def revoke_auth_token(self, env, data):
        # TODO: Blacklist old token? Immediately close websockets?
        response = {'error': 'DELETE /authenticate is not yet implemented'}
        return Response(json.dumps(response), status=501, content_type='application/json')

    def __get_user(self, username, password):
        """
        Retrieve user from the user store based upon username/password

        The hashed_password will not be returned with the value in the user
        object.

        If there is not a username/password that match return None.

        :param username:
        :param password:
        :return:
        """
        user = self._userdict.get(username)
        if user is not None:
            hashed_pass = user.get('hashed_password')
            if hashed_pass and argon2.verify(password, hashed_pass):
                usr_cpy = user.copy()
                del usr_cpy['hashed_password']
                return usr_cpy
        return None




# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

from http.cookies import SimpleCookie
import logging

from volttron.utils.context import ClientContext

try:
    import jwt
except ImportError:
    pass

from . discovery import DiscoveryInfo, DiscoveryError

# Used outside so we make it available through this file.
from . platform_web_service import PlatformWebService

_log = logging.getLogger(__name__)


class NotAuthorized(Exception):
    pass


def get_bearer(env):

    # Test if HTTP_AUTHORIZATION header is passed
    http_auth = env.get('HTTP_AUTHORIZATION')
    if http_auth:
        auth_type, bearer = http_auth.split(' ')
        if auth_type.upper() != 'BEARER':
            raise NotAuthorized("Invalid HTTP_AUTHORIZATION header passed, must be Bearer")
        else:
            return bearer
    else:
        cookiestr = env.get('HTTP_COOKIE')
        if not cookiestr:
            raise NotAuthorized()
        cookie = SimpleCookie(cookiestr)
        if 'Bearer' in cookie:
            bearer = cookie.get('Bearer').value
            return bearer
        else:
            return None


def get_user_claims(env, ssl_public_key):
    algorithm, encode_key = __get_key_and_algorithm__(env, ssl_public_key)
    bearer = get_bearer(env)
    return jwt.decode(bearer, encode_key, algorithm=algorithm)


def __get_key_and_algorithm__(env, ssl_public_key):
    publickey = env.get("WEB_PUBLIC_KEY")
    if publickey is not None or ssl_public_key is not None:
        algorithm = 'RS256'
    else:
        algorithm = 'HS256'

    if algorithm == 'HS256':
        if ClientContext.get_config_param('web-secret-key') is None:
            raise ValueError("invalid configuration detected web_secret_key must be set!")

    if algorithm == 'RS256' and ssl_public_key is None:
        encode_key = publickey
    elif algorithm == 'RS256' and ssl_public_key:
        encode_key = ssl_public_key
    else:
        encode_key = ClientContext.get_config_param('web-secret-key')

    return algorithm, encode_key


def get_user_claim_from_bearer(bearer, web_secret_key=None, tls_public_key=None):
    if web_secret_key is None and tls_public_key is None:
        raise ValueError("web_secret_key or tls_public_key must be set")
    if web_secret_key is not None and tls_public_key is not None:
        raise ValueError("web_secret_key or tls_public_key must be set not both")

    if web_secret_key is not None:
        algorithm = 'HS256'
        pubkey = web_secret_key
    else:
        algorithm = 'RS256'
        pubkey = tls_public_key
        # if isinstance(tls_public_key, str):
        #     pubkey = CertWrapper.load_cert(tls_public_key)

    claims = jwt.decode(bearer, pubkey, algorithms=algorithm)
    return claims



import functools
import os
import re
import json
from os.path import normpath, join
from gevent.timeout import Timeout
from collections import defaultdict
from typing import List, Union

from werkzeug import Response
from werkzeug.urls import url_decode

from volttron.client.known_identities import CONFIGURATION_STORE, CONTROL
from volttron.client.vip.agent.subsystems.query import Query
from volttron.utils.jsonrpc import MethodNotFound, RemoteError
from .topic_tree import DeviceTree, TopicTree
from .vui_pubsub import VUIPubsubManager


import logging
_log = logging.getLogger(__name__)


class OverrideError(Exception):
    """Error raised by driver when the user tries to set/revert point when global override is set."""
    pass


class LockError(Exception):
    """Error raised by actuator when the user does not have a device scheduled
    and tries to use methods that require exclusive access."""
    pass


def endpoint(func):
    @functools.wraps(func)
    def verify_and_dispatch(self, env, data):
        from volttron.services.web import get_bearer
        try:
            claims = self._agent.get_user_claims(get_bearer(env))
        except Exception as e:
            _log.warning(f"Unauthorized user attempted to connect to {env.get('PATH_INFO')}. Caught Exception: {e}")
            return Response(json.dumps({'error': 'Not Authorized'}), 401, content_type='app/json')

        # Only allow only users with API permissions:
        if 'vui' not in claims.get('groups'):
            _log.warning(f"Unauthorized user attempted to connect with 'vui' claim to {env.get('PATH_INFO')}.")
            return Response(json.dumps({'error': 'Not Authorized'}), 403, content_type='app/json')

        # Dispatch endpoint:
        try:
            response = func(self, env, data)
            if not response:
                message = f"Endpoint {env['REQUEST_METHOD']} {env['PATH_INFO']} is not implemented."
                return Response(json.dumps({"error": message}), status=501, content_type='application/json')
            else:
                return response
        except TimeoutError as e:
            return Response(json.dumps({'error': f'Request Timed Out: {e}'}), 504, content_type='application/json')
        except Exception as e:
            return Response(json.dumps({'error': f'Unexpected Error: {e}'}), 500, content_type='application/json')
    return verify_and_dispatch


class VUIEndpoints:
    def __init__(self, agent):
        self._agent = agent
        q = Query(self._agent.core)
        self.local_instance_name = q.query('instance-name').get(timeout=5)
        # TODO: Load active_routes from configuration. Default can just be {'vui': {'endpoint-active': False}}
        self.active_routes = {
            'vui': {
                'endpoint-active': True,
                'platforms': {
                    'endpoint-active': True,
                    'agents': {
                        'endpoint-active': True,
                        'configs': {
                            'endpoint-active': True,
                        },
                        'enabled': {
                            'endpoint-active': True,
                        },
                        'front-ends': {
                            'endpoint-active': False,
                        },
                        'health': {
                            'endpoint-active': False,
                        },
                        'pubsub': {
                            'endpoint-active': False,
                        },
                        'rpc': {
                            'endpoint-active': True,
                        },
                        'running': {
                            'endpoint-active': True,
                        },
                        'status': {
                            'endpoint-active': True,
                        },
                        'tag': {
                            'endpoint-active': True,
                        }
                    },
                    'auths': {
                        'endpoint-active': False,
                    },
                    'configs': {
                        'endpoint-active': False,
                    },
                    'devices': {
                        'endpoint-active': True,
                    },
                    'groups': {
                        'endpoint-active': False,
                    },
                    'health': {
                        'endpoint-active': False,
                    },
                    'historians': {
                        'endpoint-active': True,
                    },
                    'known-hosts': {
                        'endpoint-active': False,
                    },
                    'pubsub': {
                        'endpoint-active': True,
                    },
                    'roles': {
                        'endpoint-active': False,
                    },
                    'status': {
                        'endpoint-active': True,
                    },
                    'statistics': {
                        'endpoint-active': False,
                    }
                },
                'devices': {
                    'endpoint-active': False,
                },
                'historians': {
                    'endpoint-active': False,
                }
            }
        }
        if self.active_routes['vui']['platforms']['pubsub']:
            self.pubsub_manager = VUIPubsubManager(self._agent)

    def get_routes(self):
        """
        Returns a list of tuples with the routes for the administration endpoints
        available in it.

        :return:
        """
        # TODO: Break this up into appends to allow configuration of which endpoints are available.
        return [
            (re.compile('^/vui/?$'), 'callable', self.handle_vui_root),
            (re.compile('^/vui/platforms/?$'), 'callable', self.handle_platforms),
            (re.compile('^/vui/platforms/[^/]+/?$'), 'callable', self.handle_platforms_platform),
            (re.compile('^/vui/platforms/[^/]+/agents/?$'), 'callable', self.handle_platforms_agents),
            (re.compile('^/vui/platforms/[^/]+/agents/[^/]+/?$'), 'callable', self.handle_platforms_agents_agent),
            (re.compile('^/vui/platforms/[^/]+/agents/[^/]+/configs/?$'), 'callable', self.handle_platforms_agents_configs),
            (re.compile('^/vui/platforms/[^/]+/agents/[^/]+/configs/.*/?$'), 'callable', self.handle_platforms_agents_configs),
            (re.compile('^/vui/platforms/[^/]+/agents/[^/]+/enabled/?$'), 'callable', self.handle_platforms_agents_enabled),
            (re.compile('^/vui/platforms/[^/]+/agents/[^/]+/rpc/?$'), 'callable', self.handle_platforms_agents_rpc),
            (re.compile('^/vui/platforms/[^/]+/agents/[^/]+/rpc/[^/]+/?$'), 'callable', self.handle_platforms_agents_rpc_method),
            (re.compile('^/vui/platforms/[^/]+/agents/[^/]+/running/?$'), 'callable', self.handle_platforms_agents_running),
            (re.compile('^/vui/platforms/[^/]+/agents/[^/]+/status/?$'), 'callable', self.handle_platforms_agents_status),
            (re.compile('^/vui/platforms/[^/]+/agents/[^/]+/tag/?$'), 'callable', self.handle_platforms_agents_tag),
            (re.compile('^/vui/platforms/[^/]+/devices/?$'), 'callable', self.handle_platforms_devices),
            (re.compile('^/vui/platforms/[^/]+/devices/.*/?$'), 'callable', self.handle_platforms_devices),
            (re.compile('^/vui/platforms/[^/]+/historians/?$'), 'callable', self.handle_platforms_historians),
            (re.compile('^/vui/platforms/[^/]+/historians/[^/]+/?$'), 'callable', self.handle_platforms_historians_historian),
            (re.compile('^/vui/platforms/[^/]+/historians/[^/]+/topics/?$'), 'callable', self.handle_platforms_historians_historian_topics),
            (re.compile('^/vui/platforms/[^/]+/historians/[^/]+/topics/.*/?$'), 'callable', self.handle_platforms_historians_historian_topics),
            (re.compile('^/vui/platforms/[^/]+/pubsub/?$'), 'callable', self.handle_platforms_pubsub),
            (re.compile('^/vui/platforms/[^/]+/pubsub/.*/?$'), 'callable', self.handle_platforms_pubsub),
            (re.compile('^/vui/platforms/[^/]+/status/?$'), 'callable', self.handle_platforms_status),
            # (re.compile('^/vui/devices/?$'), 'callable', self.handle_vui_devices),
            # (re.compile('^/vui/devices/.+/?$'), 'callable', self.handle_vui_devices_topic),
            # (re.compile('^/vui/devices/hierarchy/?$'), 'callable', self.handle_vui_devices_hierarchy),
            # (re.compile('^/vui/devices/hierarchy/.+/?$'), 'callable', self.handle_vui_devices_hierarchy_topic),
            # (re.compile('^/vui/historians/?$'), 'callable', self.handle_vui_historians),
            # (re.compile('^/vui/history/?$), 'callable', self.handle_vui_history)
        ]

    @endpoint
    def handle_vui_root(self, env: dict, data: dict) -> Response:
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        if request_method == 'GET':
            response = json.dumps(self._find_active_sub_routes(['vui'], path_info=path_info))
            return Response(response, 200, content_type='application/json')

    @endpoint
    def handle_platforms(self, env: dict, data: dict) -> Response:
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        if request_method == 'GET':
            platforms = self._get_platforms()
            response = json.dumps(self._links(path_info, platforms))
            return Response(response, 200, content_type='application/json')

    @endpoint
    def handle_platforms_platform(self, env: dict, data: dict) -> Response:
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        if request_method == 'GET':
            platform = re.match('^/vui/platforms/([^/]+)/?$', path_info).groups()[0]
            if platform in self._get_platforms():
                return Response(json.dumps(self._find_active_sub_routes(['vui', 'platforms'], path_info=path_info)),
                                200, content_type='application/json')
            else:
                return Response(json.dumps({f'error': f'Unknown platform: {platform}'}),
                                400, content_type='application/json')

    @endpoint
    def handle_platforms_agents(self, env: dict, data: dict) -> Response:
        """
        Endpoints for /vui/platforms/:platform/agents/
        :param env:
        :param data:
        :return:
        """
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        query_params = url_decode(env['QUERY_STRING'])
        platform = re.match('^/vui/platforms/([^/]+)/agents/?$', path_info).groups()[0]
        if request_method == 'GET':
            include_hidden = self._to_bool(query_params.get('include-hidden', False))
            agent_state = query_params.get('agent-state', 'running')
            if agent_state not in ['running', 'installed']:
                error = {'error': f'Unknown agent-state: {agent_state} -- must be "running", "installed",'
                                  f' or "packaged". Default is "running".'}
                return Response(json.dumps(error), 400, content_type='application/json')
            if platform not in self._get_platforms():
                error = {'error': f'Unknown platform: {platform}'}
                return Response(json.dumps(error), 400, content_type='application/json')
            else:
                agents = self._get_agents(platform, agent_state, include_hidden)
                return Response(json.dumps(self._links(path_info, agents)), 200,
                                content_type='application/json')

    @endpoint
    def handle_platforms_agents_running(self, env: dict, data: dict) -> Response:
        """
        Endpoints for /vui/platforms/:platform/agents/running
        :param env:
        :param data:
        :return:
        """
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        query_params = url_decode(env['QUERY_STRING'])
        restart = self._to_bool(query_params.get('restart', 'false'))
        platform, vip_identity = re.match('^/vui/platforms/([^/]+)/agents/([^/]+)/running/?$', path_info).groups()
        uuid = self._rpc(CONTROL, 'identity_exists', vip_identity, external_platform=platform)
        if not uuid:
            return Response(json.dumps({'error': f'Agent "{vip_identity}" not found.'}), 400,
                            content_type='application/json')

        def _agent_running(vip_identity):
            peerlist = self._rpc(CONTROL, 'peerlist', external_platform=platform)
            return True if vip_identity in peerlist else False

        if request_method == 'GET':
            status = _agent_running(vip_identity)
            return Response(json.dumps({'running': status}), 200, content_type='application/json')

        elif request_method == 'PUT':
            if restart:
                self._rpc(CONTROL, 'restart_agent', uuid, external_platform=platform)
            else:
                if _agent_running(vip_identity):
                    return Response(json.dumps({'error': f'Agent: {vip_identity} is already running.'}), 400,
                                    content_type='application/json')
                self._rpc(CONTROL, 'start_agent', uuid, external_platform=platform)
            return Response(status=204)

        elif request_method == 'DELETE':
            self._rpc(CONTROL, 'stop_agent', uuid, external_platform=platform)
            return Response(status=204)

    @endpoint
    def handle_platforms_agents_agent(self, env: dict, data: dict) -> Response:
        """
        Endpoints for /vui/platforms/:platform/agents/:vip_identity/
        :param env:
        :param data:
        :return:
        """
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        platform, vip_identity = re.match('^/vui/platforms/([^/]+)/agents/([^/]+)/?$', path_info).groups()
        if request_method == 'GET':
            active_routes = self._find_active_sub_routes(['vui', 'platforms', 'agents'], path_info=path_info)
            agent_state = self._get_agent_state(platform, vip_identity)
            # If agent is not installed, only allow configs endpoint.
            if agent_state == 'not_installed' and 'configs' in active_routes['links'].keys():
                active_routes['links'] = {'configs': active_routes['links'].pop('configs')}
            # If RPC endpoint is enabled and agent is installed but not running, disallow rpc endpoint.
            elif agent_state == 'installed' and 'rpc' in active_routes['links'].keys():
                active_routes['links'].pop('rpc')
            return Response(json.dumps(active_routes), 200, content_type='application/json')

    @endpoint
    def handle_platforms_agents_configs(self, env: dict, data: dict) -> Response:
        """
                Endpoints for /vui/platforms/:platform/agents/:vip_identity/configs/:file_name
                :param env:
                :param data:
                :return:
                """
        _log.debug('VUI: In handle_platforms_agents_configs_config')
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        config_type = env.get('CONTENT_TYPE')

        no_config_name = re.match('^/vui/platforms/([^/]+)/agents/([^/]+)/configs/?$', path_info)
        if no_config_name:
            platform, vip_identity, config_name = tuple(no_config_name.groups()) + ('',)
            query_params = url_decode(env['QUERY_STRING'])
            config_name = query_params.get('config-name')
        else:
            platform, vip_identity, config_name = re.match('^/vui/platforms/([^/]+)/agents/([^/]+)/configs/(.*)/?$',
                                                           path_info).groups()
            config_name = config_name[:-1] if config_name[-1] == '/' else config_name
        if request_method == 'GET':
            try:
                if no_config_name:
                    if vip_identity != '-':
                        setting_list = self._rpc(CONFIGURATION_STORE, 'manage_list_configs', vip_identity,
                                                 external_platform=platform)
                        route_dict = self._links(path_info, setting_list)
                        return Response(json.dumps(route_dict), 200, content_type='application/json')
                    else:
                        list_of_agents = self._rpc(CONFIGURATION_STORE, 'manage_list_stores', external_platform=platform)
                        return Response(json.dumps(list_of_agents), 200, content_type='application/json')
                elif not no_config_name:
                    setting_dict = self._rpc(CONFIGURATION_STORE, 'manage_get', vip_identity, config_name,
                                             external_platform=platform)
                    return Response(json.dumps(setting_dict), 200, content_type='application/json')
            except RemoteError as e:
                return Response(json.dumps({"Error": f"{e}"}), 400, content_type='application/json')

        elif request_method == 'PUT' and (not no_config_name):
            if config_type in ['application/json', 'text/csv', 'text/plain']:
                self._insert_config(config_type, data, vip_identity, config_name, platform)
                return Response(None, 204, content_type='application/json')
            else:
                return Response(
                    json.dumps({"Error": "The configuration type can only be 'JSON', 'CSV' and 'RAW.'"}), 400,
                    content_type='application/json')

        elif request_method == 'POST' and no_config_name:
            if config_type in ['application/json', 'text/csv', 'text/plain']:
                setting_list = self._rpc(CONFIGURATION_STORE, 'manage_list_configs', vip_identity,
                                         external_platform=platform)
                if config_name in setting_list:
                    e = {'Error': f'Configuration: "{config_name}" already exists for agent: "{vip_identity}"'}
                    return Response(json.dumps(e), 409, content_type='application/json')
                self._insert_config(config_type, data, vip_identity, config_name, platform)
                response = Response(None, 201, content_type='application/json')
                response.location = f'/platforms/{platform}/agents/{vip_identity}/configs/{config_name}'
                return response
            else:
                return Response(
                    json.dumps({"Error": "The configuration type can only be 'JSON', 'CSV' and 'RAW.'"}), 400,
                    content_type='application/json')

        elif request_method == 'DELETE':
            if no_config_name:
                try:
                    self._rpc(CONFIGURATION_STORE, 'manage_delete_store', vip_identity, external_platform=platform)
                    return Response(None, 204, content_type='application/json')
                except RemoteError as e:
                    return Response(json.dumps({"Error": f"{e}"}), 400, content_type='application/json')
            else:
                try:
                    self._rpc(CONFIGURATION_STORE, 'manage_delete_config', vip_identity, config_name,
                              external_platform=platform)
                    return Response(None, 204, content_type='application/json')
                except RemoteError as e:
                    return Response(json.dumps({"Error": f"{e}"}), 400, content_type='application/json')

    @endpoint
    def handle_platforms_agents_enabled(self, env: dict, data: dict) -> Response:
        """
        Endpoints for /vui/platforms/:platform/agents/:vip_identity/enabled/
        :param env:
        :param data:
        :return:
        """
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        query_params = url_decode(env['QUERY_STRING'])
        priority = query_params.get('priority', '50')
        platform, vip_identity = re.match('^/vui/platforms/([^/]+)/agents/([^/]+)/enabled/?$', path_info).groups()

        if request_method == 'GET':
            try:
                list_of_agents = self._rpc(CONTROL, 'list_agents', external_platform=platform)
                result = next(item['priority'] for item in list_of_agents if item['identity'] == vip_identity)
                result = int(result) if result else result
                status = True if result is not None else False
                ret_val = {'status': status, 'priority': result}
                return Response(json.dumps(ret_val), 200, content_type='application/json')
            except StopIteration:
                return Response(json.dumps({'error': f'Agent "{vip_identity}" not found.'}),
                                400, content_type='application/json')
            except MethodNotFound or ValueError as e:
                return Response(json.dumps({f'error': f'For agent {vip_identity}: {e}'}),
                                400, content_type='application/json')

        elif request_method == 'PUT':
            if not priority.isdigit() or not 0 <= int(priority) < 100:
                error = {'error': f'Priority must be an integer from 0 - 99. Received: {priority}.'}
                return Response(json.dumps(error), 400, content_type='application/json')
            try:
                uuid = self._rpc(CONTROL, 'identity_exists', vip_identity, external_platform=platform)
                self._rpc(CONTROL, 'prioritize_agent', uuid, priority, external_platform=platform)
                return Response(status=204)
            except MethodNotFound or ValueError as e:
                return Response(json.dumps({f'error': f'For agent {vip_identity}: {e}'}),
                                400, content_type='application/json')

        elif request_method == 'DELETE':
            try:
                uuid = self._rpc(CONTROL, 'identity_exists', vip_identity, external_platform=platform)
                self._rpc(CONTROL, 'prioritize_agent', uuid, None, external_platform=platform)
                return Response(status=204)
            except MethodNotFound or ValueError as e:
                return Response(json.dumps({f'error': f'For agent {vip_identity}: {e}'}),
                                400, content_type='application/json')

    @endpoint
    def handle_platforms_agents_rpc(self, env: dict, data: dict) -> Response:
        """
        Endpoints for /vui/platforms/:platform/agents/:vip_identity/rpc/
        :param env:
        :param data:
        :return:
        """
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        platform, vip_identity = re.match('^/vui/platforms/([^/]+)/agents/([^/]+)/rpc/?$', path_info).groups()
        if request_method == 'GET':
            method_dict = self._rpc(vip_identity, 'inspect', external_platform=platform)
            response = self._links(path_info, method_dict.get('methods'))
            return Response(json.dumps(response), 200, content_type='application/json')

    @endpoint
    def handle_platforms_agents_rpc_method(self, env: dict, data: Union[dict, List]) -> Response:
        """
        Endpoints for /vui/platforms/:platform/agents/:vip_identity/rpc/
        :param env:
        :param data:
        :return:
        """
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        platform, vip_identity, method_name = re.match('^/vui/platforms/([^/]+)/agents/([^/]+)/rpc/([^/]+)/?$',
                                                       path_info).groups()
        if request_method == 'GET':
            try:
                method_dict = self._rpc(vip_identity, method_name + '.inspect', external_platform=platform)
            except MethodNotFound as e:
                return Response(json.dumps({f'error': f'for agent {vip_identity}: {e}'}),
                                400, content_type='application/json')
            return Response(json.dumps(method_dict), 200, content_type='application/json')

        elif request_method == 'POST':
            try:
                if type(data) is dict:
                    if 'args' in data.keys() and type(data['args']) is list:
                        args = data.pop('args')
                        result = self._rpc(vip_identity, method_name, *args, **data, external_platform=platform)
                    else:
                        result = self._rpc(vip_identity, method_name, **data, external_platform=platform)
                elif type(data) is list:
                    result = self._rpc(vip_identity, method_name, *data, external_platform=platform)
                else:
                    raise ValueError(f'Malformed message body: {data}')
            except MethodNotFound or ValueError as e:
                return Response(json.dumps({f'error': f'for agent {vip_identity}: {e}'}),
                                400, content_type='application/json')
            return Response(json.dumps(result), 200, content_type='application/json')

    @endpoint
    def handle_platforms_agents_status(self, env: dict, data: dict) -> Response:
        """
        Endpoints for /vui/platforms/:platform/agents/:vip_identity/status/
        :param env:
        :param data:
        :return:
        """
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        platform, vip_identity = re.match('^/vui/platforms/([^/]+)/agents/([^/]+)/status/?$', path_info).groups()

        if request_method == 'GET':
            try:
                status_dict = self._get_status(platform)
                our_agent = status_dict[vip_identity]
                return Response(json.dumps(our_agent), 200,
                                content_type='application/json')
            except KeyError:
                return Response(json.dumps({'error': f'Agent "{vip_identity}" not found.'}),
                                400, content_type='application/json')
            except MethodNotFound or ValueError as e:
                return Response(json.dumps({f'error': f'For agent  {e}'}),
                                400, content_type='application/json')

    @endpoint
    def handle_platforms_agents_tag(self, env: dict, data: dict) -> Response:
        """
        Endpoints for /vui/platforms/:platform/agents/:vip_identity/tag/
        :param env:
        :param data:
        :return:
        """
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        platform, vip_identity = re.match('^/vui/platforms/([^/]+)/agents/([^/]+)/tag/?$', path_info).groups()
        list_of_agents = self._rpc(CONTROL, 'list_agents', external_platform=platform)
        if request_method == 'GET':
            try:
                result = next(item['tag'] for item in list_of_agents if item['identity'] == vip_identity)
                return Response(json.dumps({'tag': f"{result}"}), 200, content_type='application/json')
            except StopIteration:
                return Response(json.dumps({'error': f"Agent '{vip_identity}' not found."}),
                                400, content_type='application/json')
            except MethodNotFound or ValueError as e:
                return Response(json.dumps({f'error': f"For agent '{vip_identity}': {e}"}),
                                400, content_type='application/json')

        elif request_method == 'PUT':
            tag = data.get('tag')
            uuid = self._rpc(CONTROL, 'identity_exists', vip_identity, external_platform=platform)
            if not uuid:
                return Response(json.dumps({'error': f"Agent '{vip_identity}' not found."}),
                                400, content_type='application/json')
            self._rpc(CONTROL, 'tag_agent', uuid, tag, external_platform=platform)
            return Response(status=204)

        elif request_method == 'DELETE':
            uuid = self._rpc(CONTROL, 'identity_exists', vip_identity, external_platform=platform)
            if not uuid:
                return Response(json.dumps({'error': f"Agent '{vip_identity}' not found."}),
                                400, content_type='application/json')
            self._rpc(CONTROL, 'tag_agent', uuid, None, external_platform=platform)
            return Response(status=204)

    @endpoint
    def handle_platforms_devices(self, env: dict, data: dict) -> Response:
        """
        Endpoints for /vui/platforms/:platform/devices/ and /vui/platforms/:platform/devices/:topic/
        :param env:
        :param data:
        :return:
        """
        def _get_allowed_write_selection(points, topic, regex, tag):
            # Query parameters:
            write_all = self._to_bool(query_params.get('write-all', False))
            confirm_values = self._to_bool(query_params.get('confirm-values', False))
            # Map of selected topics to routes:
            selection = {p.topic: f'/vui/platforms/{platform}/{p.identifier}' for p in points}
            unwritables = [p.topic for p in points if not self._to_bool(p.data['Writable'])]

            if (regex or tag or '/-' in topic or len(points) > 1) and not write_all:
                # Disallow potential use of multiple writes without explicit write-all flag:
                error_message = {
                    'error': f"Use of wildcard expressions, regex, or tags may set multiple points. "
                             f"Query must include 'write-all=true'.",
                    'requested_topic': topic,
                    'regex': regex,
                    'tag': tag,
                    'selected_points': selection
                }
                raise ValueError(json.dumps(error_message))
            elif len(unwritables) == len(points):
                raise ValueError(json.dumps({'error': 'No selected points are writable.',
                                             'unwritable_points': unwritables}))
            else:
                return confirm_values, selection, unwritables

        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        query_params = url_decode(env['QUERY_STRING'])

        tag = query_params.get('tag')
        tag = tag if tag and tag.lower() != 'null' and tag.lower() != 'none' else None
        regex = query_params.get('regex')
        regex = regex if regex and regex.lower() != 'null' and regex.lower() != 'none' else None

        no_topic = re.match('^/vui/platforms/([^/]+)/devices/?$', path_info)
        if no_topic:
            platform, topic = no_topic.groups()[0], ''
        else:
            platform, topic = re.match('^/vui/platforms/([^/]+)/devices/(.*)/?$', path_info).groups()
            topic = topic[:-1] if topic[-1] == '/' else topic

        # Resolve tags if the tag query parameter is set:
        if tag:
            try:
                tag_list = self._rpc('platform.tagging', 'get_topics_by_tags', tag, external_platform=platform)
            except Timeout as e:
                return Response(json.dumps({'error': f'Tagging Service timed out: {e}'}),
                                504, content_type='application/json')
        else:
            tag_list = None
        # Prune device tree and get nodes matching topic:
        try:
            # TODO: Should we be storing this tree to use for faster requests later? How to keep it updated?
            device_tree = DeviceTree.from_store(platform, self._rpc).prune(topic, regex, tag_list)
            topic_nodes = device_tree.get_matches(f'devices/{topic}' if topic else 'devices')
            if not topic_nodes:
                return Response(json.dumps({f'error': f'Device topic {topic} not found on platform: {platform}.'}),
                                400, content_type='application/json')
            points = device_tree.points()
        except Timeout as e:
            return Response(json.dumps({'error': f'RPC Timed Out: {e}'}), 504, content_type='application/json')

        if request_method == 'GET':
            # Query parameters:
            read_all = self._to_bool(query_params.get('read-all', False))
            return_routes = self._to_bool(query_params.get('routes', True))
            return_writability = self._to_bool(query_params.get('writability', True))
            return_values = self._to_bool(query_params.get('values', True))
            return_config = self._to_bool(query_params.get('config', False))

            try:
                if read_all or all([n.is_point() for n in topic_nodes]):
                    # Either leaf values are explicitly requested, or all nodes are already points -- Return points:
                    ret_dict = defaultdict(dict)
                    if return_values:
                        ret_values = self._rpc('platform.actuator', 'get_multiple_points',
                                               [d.topic for d in points], external_platform=platform)
                        for k, v in ret_values[0].items():
                            ret_dict[k]['value'] = v
                        for k, e in ret_values[1].items():
                            ret_dict[k]['value_error'] = e
                    for point in points:
                        if return_routes:
                            ret_dict[point.topic]['route'] = f'/vui/platforms/{platform}/{point.identifier}'
                        if return_writability:
                            ret_dict[point.topic]['writable'] = self._to_bool(point.data.get('Writable'))
                        if return_config:
                            ret_dict[point.topic]['config'] = point.data
                    return Response(json.dumps(ret_dict), 200, content_type='application/json')
                else:
                    # All topics are not complete to points and read_all=False -- return route to next segments:
                    ret_dict = {
                        'links': device_tree.get_children_dict([n.identifier for n in topic_nodes],
                                                                       replace_topic=topic,
                                                                       prefix=f'/vui/platforms/{platform}')
                    }
                    return Response(json.dumps(ret_dict), 200, content_type='application/json')

            except Timeout as e:
                return Response(json.dumps({'error': f'RPC Timed Out: {e}'}), 504, content_type='application/json')

        elif request_method == 'PUT':
            try:
                confirm_values, selected_routes, unwritables = _get_allowed_write_selection(points, topic, regex, tag)
            except ValueError as e:
                return Response(str(e), 405, content_type='application/json')

            # Set selected points:
            try:
                set_value = data.get('value')
                topics_values = [(d.topic, set_value) for d in points if d.topic not in unwritables]
                ret_errors = self._rpc('platform.actuator', 'set_multiple_points',
                                       requester_id=self._agent.core.identity, topics_values=topics_values,
                                       external_platform=platform)
                ret_dict = defaultdict(dict)
                for k in selected_routes.keys():
                    ret_dict[k]['route'] = selected_routes[k]
                    ret_dict[k]['set_error'] = ret_errors.get(k)
                    ret_dict[k]['writable'] = True if k not in unwritables else False
                if confirm_values:
                    ret_values = self._rpc('platform.actuator', 'get_multiple_points',
                                           [d.topic for d in points], external_platform=platform)
                    for k in selected_routes.keys():
                        ret_dict[k]['value'] = ret_values[0].get(k)
                        ret_dict[k]['value_check_error'] = ret_values[1].get(k)

                return Response(json.dumps(ret_dict), 200, content_type='application/json')

            except (LockError, OverrideError) as e:
                return Response(json.dumps({'error': e}), 409, content_type='application/json')
            except Timeout as e:
                return Response(json.dumps({'error': f'RPC Timed Out: {e}'}), 504, content_type='application/json')

        elif request_method == 'DELETE':
            try:
                confirm_values, selected_routes, unwritables = _get_allowed_write_selection(points, topic, regex, tag)
            except ValueError as e:
                return Response(str(e), status=405, content_type='application/json')

            # Reset selected points:
            try:
                for t_node in topic_nodes:
                    if t_node.is_device():
                        self._rpc('platform.actuator', 'revert_device', requester_id=self._agent.core.identity,
                                  topic=t_node.topic, external_platform=platform)
                    elif t_node.is_point() and t_node.topic not in unwritables:
                        self._rpc('platform.actuator', 'revert_point', requester_id=self._agent.core.identity,
                                  topic=t_node.topic, external_platform=platform)

                ret_dict = defaultdict(dict)
                for k in selected_routes.keys():
                    ret_dict[k]['route'] = selected_routes[k]
                    ret_dict[k]['writable'] = True if k not in unwritables else False

                if confirm_values:
                    ret_values = self._rpc('platform.actuator', 'get_multiple_points',
                                           [d.topic for d in points], external_platform=platform)
                    for k in selected_routes.keys():
                        ret_dict[k]['value'] = ret_values[0].get(k)
                        ret_dict[k]['value_check_error'] = ret_values[1].get(k)
                return Response(json.dumps(ret_dict), 200, content_type='application/json')

            except (LockError, OverrideError) as e:
                return Response(json.dumps({'error': e}), 409, content_type='application/json')
            except Timeout as e:
                return Response(json.dumps({'error': f'RPC Timed Out: {e}'}), 504, content_type='application/json')

        else:
            return Response(f'Endpoint {request_method} {path_info} is not implemented.',
                            status=501, content_type='text/plain')

    def handle_platforms_pubsub(self, env: dict, start_response, data: dict):
        from volttron.services.web import get_bearer  # TODO: Is this necessary, with bearer imported in decorator?
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        query_params = url_decode(env['QUERY_STRING'])
        _log.debug('VUI.handle_platforms_pubsub -- env is: ')
        _log.debug({k: str(v) for k, v in env.items()})
        _log.debug(f'HTTP_AUTHORIZATION is: {env["HTTP_AUTHORIZATION"]}')
        access_token = get_bearer(env)

        no_topic = re.match('^/vui/platforms/([^/]+)/pubsub/?$', path_info)
        if no_topic:
            platform, topic = no_topic.groups()[0], ''
        else:
            platform, topic = re.match('^/vui/platforms/([^/]+)/pubsub/(.*)/?$', path_info).groups()
            topic = topic[:-1] if topic[-1] == '/' else topic

        # GET -- For ../pubsub and /pubsub/:topic, Get routes to open web sockets for this user.
        if request_method == 'GET':
            if not topic:
                ret_dict = self.pubsub_manager.get_socket_routes(access_token, topic)
                response = Response(json.dumps(ret_dict), 200, content_type='application/json')
                return response
            else:
                ws = self.pubsub_manager.open_subscription_socket(access_token, topic)
                env['ws4py.app'] = self.pubsub_manager
                _log.debug('ENV is:')
                _log.debug(env)
                return [ws(env, start_response)]

        elif request_method == 'PUT':
            # PUT -- for ../pubsub/:topic: One-time publish to a topic.
            message = data.get('message')
            headers = data.get('headers')
            subscriber_count = self.pubsub_manager.publish(topic, headers, message)
            return Response(json.dumps(subscriber_count), 200, content_type='application/json')

        # elif request_method == 'DELETE':
        #     # DELETE -- For ../pubsub and /pubsub/:topic, Close open web sockets and subscriptions for this user.
        #     self.pubsub_manager.close_socket(access_token, topic)
        #     return Response(status=204)

    @endpoint
    def handle_platforms_historians(self, env: dict, data: dict) -> Response:
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        platform = re.match('^/vui/platforms/([^/]+)/historians/?$', path_info).groups()[0]

        if request_method == 'GET':
            agents = self._get_agents(platform)
            response = json.dumps(self._links(path_info, [agent for agent in agents if 'historian' in agent]))
            return Response(response, 200, content_type='application/json')

    @endpoint
    def handle_platforms_historians_historian(self, env: dict, data: dict) -> Response:
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        platform, vip_identity = re.match('^/vui/platforms/([^/]+)/historians/([^/]+)/?$', path_info).groups()

        if request_method == 'GET':
            links = {'links': {'topics': f'/vui/platforms/{platform}/historians/{vip_identity}/topics'}}

            return Response(json.dumps(links), 200, content_type='application/json')

    @endpoint
    def handle_platforms_historians_historian_topics(self, env: dict, data: dict) -> Response:
        """
        Endpoints for /vui/platforms/:platform/historians/topics and /vui/platforms/:platform/historians/topics/:topic/
        :param env:
        :param data:
        :return:
        """

        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        query_params = url_decode(env['QUERY_STRING'])

        # Query parameters used directly in this method.
        tag = query_params.get('tag')
        tag = tag if tag and tag.lower() != 'null' and tag.lower() != 'none' else None
        regex = query_params.get('regex')
        regex = regex if regex and regex.lower() != 'null' and regex.lower() != 'none' else None

        # Query parameters passed directly to RPC.
        start = query_params.get('start')
        end = query_params.get('end')
        skip = int(query_params.get('skip') if query_params.get('skip') else 0)
        count = query_params.get('count')
        order = query_params.get('order') if query_params.get('order') else 'FIRST_TO_LAST'
        # TODO: agg_type & agg_period not implemented, need to check response format of Aggregate Historians
        agg_type = None
        agg_period = None

        no_topic = re.match('^/vui/platforms/([^/]+)/historians/([^/]+)/topics/?$', path_info)
        if no_topic:
            platform, historian, topic = no_topic.groups()[0], no_topic.groups()[1], ''
        else:
            platform, historian, topic = re.match('^/vui/platforms/([^/]+)/historians/([^/]+)/topics/(.*)/?$',
                                                  path_info).groups()
            topic = topic[:-1] if topic[-1] == '/' else topic

        # Resolve tags if the tag query parameter is set:
        if tag:
            try:
                tag_list = self._rpc('platform.tagging', 'get_topics_by_tags', tag,
                                     external_platform=platform).get(timeout=5)
            except Timeout as e:
                return Response(json.dumps({'error': f'Tagging Service timed out: {e}'}),
                                504, content_type='application/json')
        else:
            tag_list = None
        try:
            historian_topics = self._rpc(historian, 'get_topic_list', external_platform=platform)
            historian_tree = TopicTree(historian_topics, 'historians').prune(topic, regex, tag_list)
            topic_nodes = historian_tree.get_matches(f'historians/{topic}' if topic else 'historians')

            if not topic_nodes:
                return Response(json.dumps({f'error': f'Historian topic {topic} not found on platform: {platform}.'}),
                                400, content_type='application/json')
            points = historian_tree.leaves()
        except Timeout as e:
            return Response(json.dumps({'error': f'RPC Timed Out: {e}'}), 504, content_type='application/json')

        if request_method == 'GET':
            read_all = self._to_bool(query_params.get('read-all', False))
            return_routes = self._to_bool(query_params.get('routes', True))
            return_values = self._to_bool(query_params.get('values', True))

            try:
                if read_all or all([n.is_leaf() for n in topic_nodes]):
                    # Either leaf values are explicitly requested, or all nodes are already points -- Return points:
                    ret_dict = defaultdict(dict)

                    if return_values:
                        ret_values = self._rpc(historian, 'query',
                                               [p.topic for p in points], start, end, agg_type, agg_period, skip, count,
                                               order, external_platform=platform)
                        # to match single and multiple topics query results into the same structure
                        ret_values['values'] = ret_values['values'] if isinstance(ret_values['values'], dict) else {
                            points[0].topic: ret_values['values']}
                        for k, v in ret_values['values'].items():
                            ret_dict[k]['value'] = v
                            if ret_values['metadata']:
                                ret_dict[k]['metadata'] = ret_values['metadata']
                    for point in points:
                        if return_routes:
                            ret_dict[point.topic][
                                'route'] = f'/vui/platforms/{platform}/historians/{historian}/{point.identifier}'

                    return Response(json.dumps(ret_dict), 200, content_type='application/json')

                else:
                    # All topics are not complete to points and read_all=False -- return route to next segments:
                    ret_dict = {
                        'links': historian_tree.get_children_dict([n.identifier for n in topic_nodes],
                                                                          replace_topic=f'{historian}/topics/{topic}',
                                                                          prefix=f'/vui/platforms/{platform}')}
                    return Response(json.dumps(ret_dict), 200, content_type='application/json')

            except Timeout as e:
                return Response(json.dumps({'error': f'RPC Timed Out: {e}'}), 504, content_type='application/json')

        else:
            return Response(f'Endpoint {request_method} {path_info} is not implemented.',
                            status='501 Not Implemented', content_type='text/plain')

    @endpoint
    def handle_platforms_status(self, env: dict, data: dict) -> Response:
        """
        Endpoints for /vui/platforms/:platform/status/
        :param env:
        :param data:
        :return:
        """
        path_info = env.get('PATH_INFO')
        request_method = env.get("REQUEST_METHOD")
        platform = re.match('^/vui/platforms/([^/]+)/status/?$', path_info).groups()[0]

        if request_method == 'GET':
            try:
                status_dict = self._get_status(platform)
                return Response(json.dumps(status_dict), 200,
                                content_type='application/json')
            except MethodNotFound or ValueError as e:
                return Response(json.dumps({f'error': f'For agent  {e}'}),
                                400, content_type='application/json')
        if request_method == 'DELETE':
            self._rpc(CONTROL, 'clear_status', True, external_platform=platform)
            return Response(status=204)

    def _find_active_sub_routes(self, segments: list, path_info: str = None, enclose=True) -> dict or list:
        """
        Returns active routes with constant segments at the end of the route.
                If no path_info is provided, return only a list of the keys.
        """
        route_obj = self.active_routes
        for segment in segments:
            if route_obj and route_obj.get(segment) and route_obj.get(segment).get('endpoint-active'):
                route_obj = route_obj.get(segment)
            else:
                route_obj = {}
        keys = [k for k in route_obj.keys() if k != 'endpoint-active' and route_obj[k]['endpoint-active']]
        if not path_info:
            return keys
        else:
            return self._links(path_info, keys) if enclose else self._links(path_info, keys, False)

    def _get_platforms(self):
        platforms = []
        try:
            with open(join(self._agent.core.volttron_home, 'external_platform_discovery.json')) as f:
                platforms = [platform for platform in json.load(f).keys()]
        except FileNotFoundError:
            _log.info('Did not find VOLTTRON_HOME/external_platform_discovery.json. Only local platform available.')
        except Exception as e:
            _log.warning(f'Error opening external_platform_discovery.json: {e}')
        finally:
            if self.local_instance_name not in platforms:
                platforms.insert(0, self.local_instance_name)
        return platforms

    def _get_agents(self, platform: str, agent_state: str = "running", include_hidden=False) -> List[str]:
        agent_list = self._rpc(CONTROL, 'list_agents', external_platform=platform)
        peerlist = self._rpc(CONTROL, 'peerlist', external_platform=platform)
        for agent in agent_list:
            agent['running'] = True if agent['identity'] in peerlist else False
        if include_hidden:
            for p in peerlist:
                if p not in [a['identity'] for a in agent_list]:
                    agent_list.append({'identity': p, 'running': True})
        if agent_state == 'running':
            return [a['identity'] for a in agent_list if a['running']]
        elif agent_state == 'installed':
            return [a['identity'] for a in agent_list]
        elif agent_state == 'packaged':
            return [os.path.splitext(a)[0] for a in os.listdir(f'{self._agent.core.volttron_home}/packaged')]

    def _get_agent_state(self, platform: str, vip_identity: str) -> str:
        agent_list = self._rpc(CONTROL, 'list_agents', external_platform=platform)
        installed_agents = [a['identity'] for a in agent_list]
        peerlist = self._rpc(CONTROL, 'peerlist', external_platform=platform)
        if vip_identity in installed_agents:
            return 'running' if vip_identity in peerlist else 'installed'
        else:
            return 'not_installed'

    def _get_status(self, platform: str):
        list_of_agents = self._rpc(CONTROL, 'list_agents', external_platform=platform)
        running_agents = self._rpc(CONTROL, 'status_agents', external_platform=platform)
        running_agents = {a[0]: {'pid': a[2][0], 'exit_code': a[2][1]} for a in running_agents}
        ret_dict = {}
        for la in list_of_agents:
            ra = running_agents.get(la['uuid'])
            agent_identity = la.pop('identity')
            la['running'] = True if ra and ra['exit_code'] is None else False
            la['enabled'] = True if la['priority'] else False
            la['pid'] = ra['pid'] if ra else None
            la['exit_code'] = ra['exit_code'] if ra else None
            ret_dict[agent_identity] = la
        return ret_dict

    def _insert_config(self, config_type, data, vip_identity, config_name, platform):
        config_type = re.search(r'([^\/]+$)', config_type).group() if config_type in ['application/json',
                                                                                      'text/csv'] else 'raw'
        if config_type == 'json':
            data = json.dumps(data)
        self._rpc(CONFIGURATION_STORE, 'manage_store', vip_identity, config_name, data, config_type,
                  external_platform=platform)
        return None

    def _rpc(self, vip_identity, method, *args, external_platform=None, **kwargs):
        external_platform = {'external_platform': external_platform}\
            if external_platform != self.local_instance_name else {}
        result = self._agent.vip.rpc.call(vip_identity, method, *args, **external_platform, **kwargs).get(timeout=5)
        return result

    @staticmethod
    def _links(path_info, option_segments, enclosing_dict=True):
        links = {segment: normpath('/'.join([path_info, segment])) for segment in option_segments}
        return links if not enclosing_dict else {'links': links}

    @staticmethod
    def _to_bool(values):
        values = values if type(values) is list else [values]
        bools = []
        for v in values:
            bools.append(True if str(v).lower() in ['true', 't', '1'] else False)
        if len(bools) == 1:
            return bools[0]
        elif len(bools) == 0:
            return None
        else:
            return bools



# -*- coding: utf-8 -*- {{{
# vim: set fenc=utf-8 ft=python sw=4 ts=4 sts=4 et:
#
# Copyright 2020, Battelle Memorial Institute.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# This material was prepared as an account of work sponsored by an agency of
# the United States Government. Neither the United States Government nor the
# United States Department of Energy, nor Battelle, nor any of their
# employees, nor any jurisdiction or organization that has cooperated in the
# development of these materials, makes any warranty, express or
# implied, or assumes any legal liability or responsibility for the accuracy,
# completeness, or usefulness or any information, apparatus, product,
# software, or process disclosed, or represents that its use would not infringe
# privately owned rights. Reference herein to any specific commercial product,
# process, or service by trade name, trademark, manufacturer, or otherwise
# does not necessarily constitute or imply its endorsement, recommendation, or
# favoring by the United States Government or any agency thereof, or
# Battelle Memorial Institute. The views and opinions of authors expressed
# herein do not necessarily state or reflect those of the
# United States Government or any agency thereof.
#
# PACIFIC NORTHWEST NATIONAL LABORATORY operated by
# BATTELLE for the UNITED STATES DEPARTMENT OF ENERGY
# under Contract DE-AC05-76RL01830
# }}}

import logging
import os
import re
from urllib.parse import parse_qs

from volttron.client.known_identities import PLATFORM_WEB, AUTH
from volttron.utils.jsonrpc import RemoteError

try:
    from jinja2 import Environment, FileSystemLoader, select_autoescape, TemplateNotFound
except ImportError:
    logging.getLogger().warning("Missing jinja2 library in admin_endpoints.py")

try:
    from passlib.hash import argon2
except ImportError:
    logging.getLogger().warning("Missing passlib library in admin_endpoints.py")

from watchdog_gevent import Observer
from werkzeug import Response

from volttron.utils.context import ClientContext
from volttron.utils import jsonapi
from volttron.utils.certs import Certs
from volttron.utils.filewatch import VolttronHomeFileReloader
from volttron.utils.persistance import PersistentDict


_log = logging.getLogger(__name__)


def template_env(env):
    return env['JINJA2_TEMPLATE_ENV']


class AdminEndpoints(object):

    def __init__(self, rmq_mgmt=None, ssl_public_key: bytes = None, rpc_caller=None):

        self._rpc_caller = rpc_caller
        self._rmq_mgmt = rmq_mgmt

        self._pending_auths = None
        self._denied_auths = None
        self._approved_auths = None

        if ssl_public_key is None:
            self._insecure_mode = True
        else:
            self._insecure_mode = False

        # must have a none value for when we don't have an ssl context available.
        if ssl_public_key is not None:
            if isinstance(ssl_public_key, bytes):
                self._ssl_public_key = ssl_public_key.decode('utf-8')
            elif isinstance(ssl_public_key, str):
                self._ssl_public_key = ssl_public_key

            else:
                raise ValueError("Invalid type for ssl_public_key")
        else:
            self._ssl_public_key = None

        self._userdict = None
        self.reload_userdict()

        self._observer = Observer()
        self._observer.schedule(
            VolttronHomeFileReloader("web-users.json", self.reload_userdict),
            ClientContext.get_volttron_home()
        )
        self._observer.start()

    def reload_userdict(self):
        webuserpath = os.path.join(ClientContext.get_volttron_home(), 'web-users.json')
        self._userdict = PersistentDict(webuserpath, format="json")

    def get_routes(self):
        """
        Returns a list of tuples with the routes for the administration endpoints
        available in it.

        :return:
        """
        return [
            (re.compile('^/admin.*'), 'callable', self.admin)
        ]

    def admin(self, env, data):
        if len(self._userdict) == 0:
            if env.get('REQUEST_METHOD') == 'POST':
                decoded = dict((k, v if len(v) > 1 else v[0])
                               for k, v in parse_qs(data).items())
                username = decoded.get('username')
                pass1 = decoded.get('password1')
                pass2 = decoded.get('password2')

                if pass1 == pass2 and pass1 is not None:
                    _log.debug("Setting administrator password")
                    self.add_user(username, pass1, groups=['admin'])
                    return Response('', status='302', headers={'Location': '/admin/login.html'})

            template = template_env(env).get_template('first.html')
            return Response(template.render(), content_type="text/html")

        if 'login.html' in env.get('PATH_INFO') or '/admin/' == env.get('PATH_INFO'):
            template = template_env(env).get_template('login.html')
            _log.debug("Login.html: {}".format(env.get('PATH_INFO')))
            return Response(template.render(), content_type='text/html')

        return self.verify_and_dispatch(env, data)

    def verify_and_dispatch(self, env, data):
        """ Verify that the user is an admin and dispatch

        :param env: web environment
        :param data: data associated with a web form or json/xml request data
        :return: Response object.
        """
        from ..web import get_bearer, NotAuthorized
        try:
            claims = self._rpc_caller(PLATFORM_WEB, 'get_user_claims', get_bearer(env)).get()
        except NotAuthorized:
            _log.error("Unauthorized user attempted to connect to {}".format(env.get('PATH_INFO')))
            return Response('<h1>Unauthorized User</h1>', status="401 Unauthorized")
        except RemoteError as e:
            if "ExpiredSignatureError" in e.exc_info["exc_type"]:
                _log.warning("Access token has expired! Please re-login to renew.")
                template = template_env(env).get_template('login.html')
                _log.debug("Login.html: {}".format(env.get('PATH_INFO')))
                return Response(template.render(), content_type='text/html')
            else:
                _log.error(e)

        # Make sure we have only admins for viewing this.
        if 'admin' not in claims.get('groups'):
            return Response('<h1>Unauthorized User</h1>', status="401 Unauthorized")

        path_info = env.get('PATH_INFO')
        if path_info.startswith('/admin/api/'):
            return self.__api_endpoint(path_info[len('/admin/api/'):], data)

        if path_info.endswith('html'):
            page = path_info.split('/')[-1]
            try:
                template = template_env(env).get_template(page)
            except TemplateNotFound:
                return Response("<h1>404 Not Found</h1>", status="404 Not Found")

            if page == 'pending_auth_reqs.html':
                try:
                    self._pending_auths = self._rpc_caller.call(AUTH, 'get_authorization_pending').get(timeout=2)
                    self._denied_auths = self._rpc_caller.call(AUTH, 'get_authorization_denied').get(timeout=2)
                    self._approved_auths = self._rpc_caller.call(AUTH, 'get_authorization_approved').get(timeout=2)
                except TimeoutError:
                    self._pending_auths = []
                    self._denied_auths = []
                    self._approved_auths = []
                # When messagebus is rmq, include pending csrs in the output pending_auth_reqs.html page
                if self._rmq_mgmt is not None:
                    html = template.render(csrs=self._rpc_caller.call(AUTH, 'get_pending_csrs').get(timeout=4),
                                           auths=self._pending_auths,
                                           denied_auths=self._denied_auths,
                                           approved_auths=self._approved_auths)
                else:
                    html = template.render(auths=self._pending_auths,
                                           denied_auths=self._denied_auths,
                                           approved_auths=self._approved_auths)
            else:
                # A template with no params.
                html = template.render()

            return Response(html, content_type="text/html")

        template = template_env(env).get_template('index.html')
        resp = template.render()
        return Response(resp)

    def __api_endpoint(self, endpoint, data):
        _log.debug("Doing admin endpoint {}".format(endpoint))
        if endpoint == 'certs':
            response = self.__cert_list_api()
        elif endpoint == 'pending_csrs':
            response = self.__pending_csrs_api()
        elif endpoint.startswith('approve_csr/'):
            response = self.__approve_csr_api(endpoint.split('/')[1])
        elif endpoint.startswith('deny_csr/'):
            response = self.__deny_csr_api(endpoint.split('/')[1])
        elif endpoint.startswith('delete_csr/'):
            response = self.__delete_csr_api(endpoint.split('/')[1])
        elif endpoint.startswith('approve_credential/'):
            response = self.__approve_credential_api(endpoint.split('/')[1])
        elif endpoint.startswith('deny_credential/'):
            response = self.__deny_credential_api(endpoint.split('/')[1])
        elif endpoint.startswith('delete_credential/'):
            response = self.__delete_credential_api(endpoint.split('/')[1])
        else:
            response = Response('{"status": "Unknown endpoint {}"}'.format(endpoint),
                                content_type="application/json")
        return response

    def __approve_csr_api(self, common_name):
        try:
            _log.debug("Creating cert and permissions for user: {}".format(common_name))
            self._rpc_caller.call(AUTH, 'approve_authorization_failure', common_name).wait(timeout=4)
            data = dict(status=self._rpc_caller.call(AUTH, "get_pending_csr_status", common_name).get(timeout=2),
                        cert=self._rpc_caller.call(AUTH, "get_pending_csr_cert", common_name).get(timeout=2))
        except ValueError as e:
            data = dict(status="ERROR", message=e.message)

        except TimeoutError as e:
            data = dict(status="ERROR", message=e.message)

        return Response(jsonapi.dumps(data), content_type="application/json")

    def __deny_csr_api(self, common_name):
        try:
            self._rpc_caller.call(AUTH, 'deny_authorization_failure', common_name).wait(timeout=2)
            data = dict(status="DENIED",
                        message="The administrator has denied the request")
        except ValueError as e:
            data = dict(status="ERROR", message=e.message)

        except TimeoutError as e:
            data = dict(status="ERROR", message=e.message)

        return Response(jsonapi.dumps(data), content_type="application/json")

    def __delete_csr_api(self, common_name):
        try:
            self._rpc_caller.call(AUTH, 'delete_authorization_failure', common_name).wait(timeout=2)
            data = dict(status="DELETED",
                        message="The administrator has denied the request")
        except ValueError as e:
            data = dict(status="ERROR", message=e.message)

        except TimeoutError as e:
            data = dict(status="ERROR", message=e.message)

        return Response(jsonapi.dumps(data), content_type="application/json")

    def __pending_csrs_api(self):
        try:
            data = self._rpc_caller.call(AUTH, 'get_pending_csrs').get(timeout=4)

        except TimeoutError as e:
            data = dict(status="ERROR", message=e.message)

        return Response(jsonapi.dumps(data), content_type="application/json")

    def __cert_list_api(self):

        try:
            data = [dict(common_name=x.common_name) for x in
                    self._rpc_caller.call(AUTH, "get_all_pending_csr_subjects").get(timeout=2)]

        except TimeoutError as e:
            data = dict(status="ERROR", message=e.message)

        return Response(jsonapi.dumps(data), content_type="application/json")

    def __approve_credential_api(self, user_id):
        try:
            _log.debug("Creating credential and permissions for user: {}".format(user_id))
            self._rpc_caller.call(AUTH, 'approve_authorization_failure', user_id).wait(timeout=4)
            data = dict(status='APPROVED',
                        message="The administrator has approved the request")
        except ValueError as e:
            data = dict(status="ERROR", message=e.message)

        except TimeoutError as e:
            data = dict(status="ERROR", message=e.message)

        return Response(jsonapi.dumps(data), content_type="application/json")

    def __deny_credential_api(self, user_id):
        try:
            self._rpc_caller.call(AUTH, 'deny_authorization_failure', user_id).wait(timeout=2)
            data = dict(status="DENIED",
                        message="The administrator has denied the request")
        except ValueError as e:
            data = dict(status="ERROR", message=e.message)

        except TimeoutError as e:
            data = dict(status="ERROR", message=e.message)

        return Response(jsonapi.dumps(data), content_type="application/json")

    def __delete_credential_api(self, user_id):
        try:
            self._rpc_caller.call(AUTH, 'delete_authorization_failure', user_id).wait(timeout=2)
            data = dict(status="DELETED",
                        message="The administrator has denied the request")
        except ValueError as e:
            data = dict(status="ERROR", message=e.message)

        except TimeoutError as e:
            data = dict(status="ERROR", message=e.message)

        return Response(jsonapi.dumps(data), content_type="application/json")

    def add_user(self, username, unencrypted_pw, groups=None, overwrite=False):
        if self._userdict.get(username) and not overwrite:
            raise ValueError(f"The user {username} is already present and overwrite not set to True")
        if groups is None:
            groups = []
        hashed_pass = argon2.hash(unencrypted_pw)
        self._userdict[username] = dict(
            hashed_password=hashed_pass,
            groups=groups
        )

        self._userdict.sync()



from typing import Union, Iterable
from treelib import Tree, Node
from treelib.exceptions import DuplicatedNodeIdError, NodeIDAbsentError
from collections import defaultdict

from volttron.client.known_identities import CONFIGURATION_STORE

import re
from os.path import normpath

import logging
_log = logging.getLogger(__name__)


class TopicNode(Node):
    def __init__(self, tag=None, identifier=None, expanded=True, data=None, segment_type='TOPIC_SEGMENT', topic=''):
        super(TopicNode, self).__init__(tag, identifier, expanded, data)
        self.segment_type = segment_type
        self.topic = topic

    def is_segment(self):
        return True if self.segment_type == 'TOPIC_SEGMENT' else False


class TopicTree(Tree):
    def __init__(self, topic_list=None, root_name='root', node_class=None, *args, **kwargs):
        node_class = node_class if node_class else TopicNode
        super(TopicTree, self).__init__(node_class=node_class, *args, **kwargs)
        if topic_list:
            self._from_topic_list(topic_list, root_name)
        else:
            self.create_node(root_name, root_name).segment_type = 'TOPIC_ROOT'

    def _from_topic_list(self, topic_list, root_name):
        tops = [t.split('/') for t in topic_list]
        if all([top[0] == root_name for top in tops]):
            [top.pop(0) for top in tops]
        self.create_node(root_name, root_name).segment_type = 'TOPIC_ROOT'
        for top in tops:
            parent = root_name
            for segment in top:
                nid = '/'.join([parent, segment])
                try:
                    self.create_node(segment, nid, parent=parent)
                except DuplicatedNodeIdError:
                    pass
                parent = nid

    def add_node(self, node, parent=None):
        super(TopicTree, self).add_node(node, parent)
        node.topic = node.identifier[(len(self.root) + 1):]
        return node

    # TODO: Should this actually be get_child_topics() where topics or routes are returned with wildcards?
    def get_children_dict(self, sub_root_node_id: Union[list, str], include_root: bool = True,
                          prefix: str = '', replace_topic: str = None) -> dict:
        sub_root_node_id = sub_root_node_id if type(sub_root_node_id) is list else [sub_root_node_id]
        level_dict = defaultdict(set)
        for r_id in sub_root_node_id:
            for d in self.children(r_id):
                try:
                    if replace_topic:
                        if include_root:
                            level_dict[d.tag].add('/'.join([self.root, replace_topic, d.tag]))
                        else:
                            level_dict[d.tag].add('/'.join([replace_topic, d.tag]))
                    else:
                        if include_root:
                            level_dict[d.tag].add(d.identifier)
                        else:
                            level_dict[d.tag].add(d.identifier.split('/', 1)[1])
                except NodeIDAbsentError as e:
                    return {}
        ret_dict = {}
        for k, s in level_dict.items():
            if len(s) > 1:
                ret_dict[k] = sorted([normpath('/'.join([prefix, v])) for v in s])
            else:
                ret_dict[k] = normpath('/'.join([prefix, s.pop()]))
        return ret_dict

    def prune(self, topic_pattern: str = None, regex: str = None, exact_matches: Iterable = None, *args, **kwargs):
        if topic_pattern:
            pattern = re.compile(topic_pattern.replace('-', '[^/]+') + '(/|$)')
            nids = [n.identifier for n in self.filter_nodes(lambda x: pattern.search(x.identifier))]
        else:
            nids = list(self.expand_tree())
        if regex:
            regex = re.compile(regex)
            nids = [n for n in nids if regex.search(n)]
        if exact_matches:
            nids = [n for n in nids if n in exact_matches]
        pruned = self.__class__(topic_list=nids, root_name=self.root, *args, **kwargs)
        for nid in [n.identifier for n in pruned.all_nodes()]:
            old = self.get_node(nid)
            pruned.update_node(nid, data=old.data, segment_type=old.segment_type)
        return pruned

    def get_matches(self, topic, return_nodes=True):
        pattern = topic.replace('-', '[^/]+') + '$'
        nodes = self.filter_nodes(lambda x: re.match(pattern, x.identifier))
        if return_nodes:
            return list(nodes)
        else:
            return [n.identifier for n in nodes]


class DeviceNode(TopicNode):
    def __init__(self, tag=None, identifier=None, expanded=True, data=None, segment_type='TOPIC_SEGMENT', topic=''):
        super(DeviceNode, self).__init__(tag, identifier, expanded, data, segment_type, topic)

    def is_point(self):
        return True if self.segment_type == 'POINT' else False

    def is_device(self):
        return True if self.segment_type == 'DEVICE' else False


class DeviceTree(TopicTree):
    def __init__(self, topic_list=None, root_name='devices', assume_full_topics=False,  *args, **kwargs):
        super(DeviceTree, self).__init__(topic_list=topic_list, root_name=root_name, node_class=DeviceNode,
                                         *args, **kwargs)
        if assume_full_topics:
            for n in self.leaves():
                n.segment_type = 'POINT'
            for n in [self.parent(l.identifier) for l in self.leaves()]:
                n.segment_type = 'DEVICE'

    def points(self, nid=None):
        if nid is None:
            points = [n for n in self._nodes.values() if n.is_point()]
        else:
            points = [self[n] for n in self.expand_tree(nid) if self[n].is_point()]
        return points

    def devices(self, nid=None):
        if nid is None:
            points = [n for n in self._nodes.values() if n.is_device()]
        else:
            points = [self[n] for n in self.expand_tree(nid) if self[n].is_device()]
        return points

    # TODO: Getting points requires getting device config, using it to find the registry config,
    #  and then parsing that. There is not a method in config.store, nor in the platform.driver for
    #  getting a completed configuration. The configuration is only fully assembled in the subsystem's
    #  _initial_update method called when the agent itself calls get_configs at startup. There does not
    #  seem to be an equivalent management method, and the code for this is in the agent subsystem
    #  rather than the service (though it is reached through the service, oddly...
    @classmethod
    def from_store(cls, platform, rpc_caller):
        # TODO: Duplicate logic for external_platform check from VUIEndpoints to remove reference to it from here.
        kwargs = {'external_platform': platform} if 'VUIEndpoints' in rpc_caller.__repr__() else {}
        devices = rpc_caller(CONFIGURATION_STORE, 'manage_list_configs', 'platform.driver', **kwargs)
        devices = devices if kwargs else devices.get(timeout=5)
        devices = [d for d in devices if re.match('^devices/.*', d)]
        device_tree = cls(devices)
        for d in devices:
            dev_config = rpc_caller(CONFIGURATION_STORE, 'manage_get', 'platform.driver', d, raw=False, **kwargs)
            # TODO: If not AsyncResponse instead of if kwargs
            dev_config = dev_config if kwargs else dev_config.get(timeout=5)
            reg_cfg_name = dev_config.pop('registry_config')[len('config://'):]
            device_tree.update_node(d, data=dev_config, segment_type='DEVICE')
            registry_config = rpc_caller('config.store', 'manage_get', 'platform.driver',
                                         f'{reg_cfg_name}', raw=False, **kwargs)
            registry_config = registry_config if kwargs else registry_config.get(timeout=5)
            for pnt in registry_config:
                point_name = pnt.pop('Volttron Point Name')
                n = device_tree.create_node(point_name, f"{d}/{point_name}", parent=d, data=pnt)
                n.segment_type = 'POINT'
        return device_tree



import logging

from ws4py.websocket import WebSocket

_log = logging.getLogger(__name__)


class VolttronWebSocket(WebSocket):

    def __init__(self, *args, **kwargs):
        super(VolttronWebSocket, self).__init__(*args, **kwargs)
        _log = logging.getLogger(self.__class__.__name__)

    def _get_identity_and_endpoint(self):
        identity = self.environ['identity']
        endpoint = self.environ['PATH_INFO']
        return identity, endpoint

    def opened(self):
        _log.info('Socket opened')
        app = self.environ['ws4py.app']
        identity, endpoint = self._get_identity_and_endpoint()
        app.client_opened(self, endpoint, identity)

    def received_message(self, m):
        # self.clients is set from within the server
        # and holds the list of all connected servers
        # we can dispatch to
        _log.debug('Socket received message: {}'.format(m))
        app = self.environ['ws4py.app']
        identity, endpoint = self._get_identity_and_endpoint()
        app.client_received(endpoint, m)

    def closed(self, code, reason="A client left the room without a proper explanation."):
        _log.info('Socket closed!')
        app = self.environ.pop('ws4py.app')
        identity, endpoint = self._get_identity_and_endpoint()
        app.client_closed(self, endpoint, identity, reason)



import re
import logging
import weakref

from werkzeug import Response

from volttron.utils import jsonapi
from volttron.utils.certs import Certs
from volttron.utils.context import ClientContext

_log = logging.getLogger(__name__)


class CSREndpoints(object):

    def __init__(self, core):
        self._core = weakref.ref(core)
        self._certs = Certs()
        self._auto_allow_csr = False

    @property
    def auto_allow_csr(self):
        return self._auto_allow_csr

    @auto_allow_csr.setter
    def auto_allow_csr(self, value):
        self._auto_allow_csr = value

    def get_routes(self):
        """
        Returns a list of tuples with the routes for authentication.

        Tuple should have the following:

            - regular expression for calling the endpoint
            - 'callable' keyword specifying that a method is being specified
            - the method that should be used to call when the regular expression matches

        code:

            return [
                (re.compile('^/csr/request_new$'), 'callable', self._csr_request_new)
            ]

        :return:
        """
        return [
            (re.compile('^/csr/request_new$'), 'callable', self._csr_request_new)
        ]

    def _csr_request_new(self, env, data):
        _log.debug("New csr request")
        if not isinstance(data, dict):
            try:
                request_data = jsonapi.loads(data)
            except:
                _log.error("Invalid data for csr request.  Must be json serializable")
                return Response()
        else:
            request_data = data.copy()

        csr = request_data.get('csr').encode("utf-8")
        identity = self._certs.get_csr_common_name(csr)

        # The identity must start with the current instances name or it is a failure.
        if not identity.startswith(ClientContext.get_instance_name() + "."):
            json_response = dict(status="ERROR",
                                 message="CSR must start with instance name: {}".format(
                                     ClientContext.get_instance_name()))
            Response(jsonapi.dumps(json_response),
                     content_type='application/json',
                     headers={'Content-type': 'application/json'})

        csr_file = self._certs.save_pending_csr_request(env.get('REMOTE_ADDR'), identity, csr)

        if self._auto_allow_csr:
            _log.debug("Creating cert and permissions for user: {}".format(identity))
            status = self._certs.get_csr_status(identity)
            json_response = dict(status=status)
            if status == 'APPROVED':
                try:
                    json_response['cert'] = self._certs.get_cert_from_csr(identity)
                except Exception as e:
                    _log.error(f"Exception getting cert from csr {e}")

            else:
                try:
                    cert = self._certs.approve_csr(identity)
                    #permissions = self._core().rmq_mgmt.get_default_permissions(identity)
                    _log.debug(f"CREATING NEW RMQ USER: {identity}")
                    permissions = dict(configure=".*", read=".*",
                                       write=".*")
                    self._core().rmq_mgmt.create_user_with_permissions(identity,
                                                                       permissions,
                                                                       True)
                    json_response = dict(
                        status="SUCCESSFUL",
                        cert=cert
                    )
                except BaseException as e:
                    _log.error(f"Exception in approving csr/creating user in auto_allow_csr mode: {e}")
        else:

            status = self._certs.get_csr_status(identity)
            cert = self._certs.get_cert_from_csr(identity)

            json_response = dict(status=status)
            if status == "APPROVED":
                json_response['cert'] = self._certs.get_cert_from_csr(identity)
            elif status == "PENDING":
                json_response['message'] = "The request is pending admininstrator approval."
            elif status == "DENIED":
                json_response['message'] = "The request has been denied by the administrator."
            elif status == "UNKNOWN":
                json_response['message'] = "An unknown common name was specified to the server {}".format(identity)
            else:
                json_response['message'] = "An unkonwn error has occured during the respons phase"

        response = None
        try:
            if json_response.get('cert', None):
                json_response['cert'] = json_response['cert'].decode('utf-8')
            response = Response(jsonapi.dumps(json_response),
                     content_type='application/json',
                     headers={'Content-type': 'application/json'})
        except BaseException as e:
            _log.error(f"Exception creating Response {e}")
        return response



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['volttron', 'volttron.services.web']

package_data = \
{'': ['*'],
 'volttron.services.web': ['static/*',
                           'static/js/*',
                           'static/specs/*',
                           'templates/*']}

install_requires = \
['PyJWT==1.7.1',
 'argon2-cffi>=21.3.0,<22.0.0',
 'jinja2-cli>=0.7.0',
 'passlib>=1.7.4,<2.0.0',
 'requests>=2.28.1,<3.0.0',
 'treelib>=1.6.1',
 'volttron>=10.0.1a43,<11.0',
 'werkzeug>=2.1.2',
 'ws4py>=0.5.1']

setup_kwargs = {
    'name': 'volttron-lib-web',
    'version': '0.1.1a2',
    'description': '',
    'long_description': 'This package provides web services for the VOLTTRON™ platform.\nThis includes a RESTful API for interacting with the platform\nand utility pages for administration and certificate management.\nThis library cannot be installed as a VOLTTRON agent is.\nRather, it must be installed as a python package.\n\n[![Run Pytests](https://github.com/eclipse-volttron/volttron-lib-web/actions/workflows/run-test.yml/badge.svg)](https://github.com/eclipse-volttron/volttron-lib-web/actions/workflows/run-test.yml)\n[![pypi version](https://img.shields.io/pypi/v/volttron.svg)](https://pypi.org/project/volttron-core/)\n\n## Requirements\n* python >=3.8\n* volttron >= 10.0\n\n## Installation with Web\nThis library can be installed using pip:\n\n```bash\n> pip install volttron-lib-web\n```\n\nOnce the library is installed, VOLTTRON will not be able to start until the\nweb service is configured. Configurations for services, including this, reside\nin a service_config.yml file in the VOLTTRON_HOME directory\n(by default ~/.volttron/service_config.yml).\nIf this file does not already exist, create it. To configure the web service,\ninclude the following:\n\n```yaml\nvolttron.services.web:\n  enabled: true\n  kwargs:\n    bind_web_address: http://127.0.0.156:5467\n    web_secret_key: some_string # If not using SSL.\n    web_ssl_cert: /path/to/certificate # Path to the SSL certificate to be used by the web service. \n    web_ssl_key: /path/to/key # Path to the SSL secret key file used by web service.\n```\n\nIf using SSL, both web_ssl_certificate and web_ssl_key are required\nand web_secret_key should not be included. If SSL is not desired,\nprovide a web_secret_key instead and remove the lines for the web_ssl_cert\nand web_ssl_key.\n\nFull VOLTTRON documentation is available at [VOLTTRON Readthedocs](https://volttron.readthedocs.io)\n\n## Development\n\nPlease see the following for contributing guidelines [contributing](https://github.com/eclipse-volttron/volttron-core/blob/develop/CONTRIBUTING.md).\n\nPlease see the following helpful guide about [developing modular VOLTTRON agents](https://github.com/eclipse-volttron/volttron-core/blob/develop/DEVELOPING_ON_MODULAR.md)\n\n# Disclaimer Notice\n\nThis material was prepared as an account of work sponsored by an agency of the\nUnited States Government.  Neither the United States Government nor the United\nStates Department of Energy, nor Battelle, nor any of their employees, nor any\njurisdiction or organization that has cooperated in the development of these\nmaterials, makes any warranty, express or implied, or assumes any legal\nliability or responsibility for the accuracy, completeness, or usefulness or any\ninformation, apparatus, product, software, or process disclosed, or represents\nthat its use would not infringe privately owned rights.\n\nReference herein to any specific commercial product, process, or service by\ntrade name, trademark, manufacturer, or otherwise does not necessarily\nconstitute or imply its endorsement, recommendation, or favoring by the United\nStates Government or any agency thereof, or Battelle Memorial Institute. The\nviews and opinions of authors expressed herein do not necessarily state or\nreflect those of the United States Government or any agency thereof.',
    'author': 'Craig Allwardt',
    'author_email': 'craig.allwardt@pnnl.gov',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://volttron.readthedocs.io',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
